#
# PySNMP MIB module IF-MIB (http://snmplabs.com/pysmi)
# Produced by pysmi-0.3.4 at Wed Aug 23 11:18:23 2023
# Using Python version 3.11.4 (main, Jun  6 2023, 22:16:46) [GCC 12.2.0]
#
Integer, OctetString, ObjectIdentifier = mibBuilder.importSymbols("ASN1", "Integer", "OctetString", "ObjectIdentifier")
NamedValues, = mibBuilder.importSymbols("ASN1-ENUMERATION", "NamedValues")
ValueSizeConstraint, SingleValueConstraint, ConstraintsIntersection, ConstraintsUnion, ValueRangeConstraint = mibBuilder.importSymbols("ASN1-REFINEMENT", "ValueSizeConstraint", "SingleValueConstraint", "ConstraintsIntersection", "ConstraintsUnion", "ValueRangeConstraint")
IANAifType, = mibBuilder.importSymbols("IANAifType-MIB", "IANAifType")
NotificationGroup, ObjectGroup, ModuleCompliance = mibBuilder.importSymbols("SNMPv2-CONF", "NotificationGroup", "ObjectGroup", "ModuleCompliance")
snmpTraps, = mibBuilder.importSymbols("SNMPv2-MIB", "snmpTraps")
NotificationType, MibScalar, MibTable, MibTableRow, MibTableColumn, mib_2, iso, ObjectIdentity, Integer32, Gauge32, ModuleIdentity, MibIdentifier, Unsigned32, TimeTicks, Counter64, Counter32, Bits, IpAddress = mibBuilder.importSymbols("SNMPv2-SMI", "NotificationType", "MibScalar", "MibTable", "MibTableRow", "MibTableColumn", "mib-2", "iso", "ObjectIdentity", "Integer32", "Gauge32", "ModuleIdentity", "MibIdentifier", "Unsigned32", "TimeTicks", "Counter64", "Counter32", "Bits", "IpAddress")
TextualConvention, PhysAddress, TruthValue, TimeStamp, RowStatus, DisplayString, TestAndIncr, AutonomousType = mibBuilder.importSymbols("SNMPv2-TC", "TextualConvention", "PhysAddress", "TruthValue", "TimeStamp", "RowStatus", "DisplayString", "TestAndIncr", "AutonomousType")
ifMIB = ModuleIdentity((1, 3, 6, 1, 2, 1, 31))
ifMIB.setRevisions(('2000-06-14 00:00', '1996-02-28 21:55', '1993-11-08 21:55',))
if mibBuilder.loadTexts: ifMIB.setLastUpdated('200006140000Z')
if mibBuilder.loadTexts: ifMIB.setOrganization('IETF Interfaces MIB Working Group')
ifMIBObjects = MibIdentifier((1, 3, 6, 1, 2, 1, 31, 1))
interfaces = MibIdentifier((1, 3, 6, 1, 2, 1, 2))
class OwnerString(TextualConvention, OctetString):
    status = 'deprecated'
    displayHint = '255a'
    subtypeSpec = OctetString.subtypeSpec + ValueSizeConstraint(0, 255)

class InterfaceIndex(TextualConvention, Integer32):
    status = 'current'
    displayHint = 'd'
    subtypeSpec = Integer32.subtypeSpec + ValueRangeConstraint(1, 2147483647)

class InterfaceIndexOrZero(TextualConvention, Integer32):
    status = 'current'
    displayHint = 'd'
    subtypeSpec = Integer32.subtypeSpec + ValueRangeConstraint(0, 2147483647)

ifNumber = MibScalar((1, 3, 6, 1, 2, 1, 2, 1), Integer32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ifNumber.setStatus('current')
ifTableLastChange = MibScalar((1, 3, 6, 1, 2, 1, 31, 1, 5), TimeTicks()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ifTableLastChange.setStatus('current')
ifTable = MibTable((1, 3, 6, 1, 2, 1, 2, 2), )
if mibBuilder.loadTexts: ifTable.setStatus('current')
ifEntry = MibTableRow((1, 3, 6, 1, 2, 1, 2, 2, 1), ).setIndexNames((0, "IF-MIB", "ifIndex"))
if mibBuilder.loadTexts: ifEntry.setStatus('current')
ifIndex = MibTableColumn((1, 3, 6, 1, 2, 1, 2, 2, 1, 1), InterfaceIndex()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ifIndex.setStatus('current')
ifDescr = MibTableColumn((1, 3, 6, 1, 2, 1, 2, 2, 1, 2), DisplayString().subtype(subtypeSpec=ValueSizeConstraint(0, 255))).setMaxAccess("readonly")
if mibBuilder.loadTexts: ifDescr.setStatus('current')
ifType = MibTableColumn((1, 3, 6, 1, 2, 1, 2, 2, 1, 3), IANAifType()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ifType.setStatus('current')
ifMtu = MibTableColumn((1, 3, 6, 1, 2, 1, 2, 2, 1, 4), Integer32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ifMtu.setStatus('current')
ifSpeed = MibTableColumn((1, 3, 6, 1, 2, 1, 2, 2, 1, 5), Gauge32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ifSpeed.setStatus('current')
ifPhysAddress = MibTableColumn((1, 3, 6, 1, 2, 1, 2, 2, 1, 6), PhysAddress()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ifPhysAddress.setStatus('current')
ifAdminStatus = MibTableColumn((1, 3, 6, 1, 2, 1, 2, 2, 1, 7), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(1, 2, 3))).clone(namedValues=NamedValues(("up", 1), ("down", 2), ("testing", 3)))).setMaxAccess("readwrite")
if mibBuilder.loadTexts: ifAdminStatus.setStatus('current')
ifOperStatus = MibTableColumn((1, 3, 6, 1, 2, 1, 2, 2, 1, 8), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(1, 2, 3, 4, 5, 6, 7))).clone(namedValues=NamedValues(("up", 1), ("down", 2), ("testing", 3), ("unknown", 4), ("dormant", 5), ("notPresent", 6), ("lowerLayerDown", 7)))).setMaxAccess("readonly")
if mibBuilder.loadTexts: ifOperStatus.setStatus('current')
ifLastChange = MibTableColumn((1, 3, 6, 1, 2, 1, 2, 2, 1, 9), TimeTicks()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ifLastChange.setStatus('current')
ifInOctets = MibTableColumn((1, 3, 6, 1, 2, 1, 2, 2, 1, 10), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ifInOctets.setStatus('current')
ifInUcastPkts = MibTableColumn((1, 3, 6, 1, 2, 1, 2, 2, 1, 11), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ifInUcastPkts.setStatus('current')
ifInNUcastPkts = MibTableColumn((1, 3, 6, 1, 2, 1, 2, 2, 1, 12), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ifInNUcastPkts.setStatus('deprecated')
ifInDiscards = MibTableColumn((1, 3, 6, 1, 2, 1, 2, 2, 1, 13), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ifInDiscards.setStatus('current')
ifInErrors = MibTableColumn((1, 3, 6, 1, 2, 1, 2, 2, 1, 14), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ifInErrors.setStatus('current')
ifInUnknownProtos = MibTableColumn((1, 3, 6, 1, 2, 1, 2, 2, 1, 15), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ifInUnknownProtos.setStatus('current')
ifOutOctets = MibTableColumn((1, 3, 6, 1, 2, 1, 2, 2, 1, 16), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ifOutOctets.setStatus('current')
ifOutUcastPkts = MibTableColumn((1, 3, 6, 1, 2, 1, 2, 2, 1, 17), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ifOutUcastPkts.setStatus('current')
ifOutNUcastPkts = MibTableColumn((1, 3, 6, 1, 2, 1, 2, 2, 1, 18), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ifOutNUcastPkts.setStatus('deprecated')
ifOutDiscards = MibTableColumn((1, 3, 6, 1, 2, 1, 2, 2, 1, 19), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ifOutDiscards.setStatus('current')
ifOutErrors = MibTableColumn((1, 3, 6, 1, 2, 1, 2, 2, 1, 20), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ifOutErrors.setStatus('current')
ifOutQLen = MibTableColumn((1, 3, 6, 1, 2, 1, 2, 2, 1, 21), Gauge32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ifOutQLen.setStatus('deprecated')
ifSpecific = MibTableColumn((1, 3, 6, 1, 2, 1, 2, 2, 1, 22), ObjectIdentifier()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ifSpecific.setStatus('deprecated')
ifXTable = MibTable((1, 3, 6, 1, 2, 1, 31, 1, 1), )
if mibBuilder.loadTexts: ifXTable.setStatus('current')
ifXEntry = MibTableRow((1, 3, 6, 1, 2, 1, 31, 1, 1, 1), )
ifEntry.registerAugmentions(("IF-MIB", "ifXEntry"))
ifXEntry.setIndexNames(*ifEntry.getIndexNames())
if mibBuilder.loadTexts: ifXEntry.setStatus('current')
ifName = MibTableColumn((1, 3, 6, 1, 2, 1, 31, 1, 1, 1, 1), DisplayString()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ifName.setStatus('current')
ifInMulticastPkts = MibTableColumn((1, 3, 6, 1, 2, 1, 31, 1, 1, 1, 2), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ifInMulticastPkts.setStatus('current')
ifInBroadcastPkts = MibTableColumn((1, 3, 6, 1, 2, 1, 31, 1, 1, 1, 3), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ifInBroadcastPkts.setStatus('current')
ifOutMulticastPkts = MibTableColumn((1, 3, 6, 1, 2, 1, 31, 1, 1, 1, 4), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ifOutMulticastPkts.setStatus('current')
ifOutBroadcastPkts = MibTableColumn((1, 3, 6, 1, 2, 1, 31, 1, 1, 1, 5), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ifOutBroadcastPkts.setStatus('current')
ifHCInOctets = MibTableColumn((1, 3, 6, 1, 2, 1, 31, 1, 1, 1, 6), Counter64()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ifHCInOctets.setStatus('current')
ifHCInUcastPkts = MibTableColumn((1, 3, 6, 1, 2, 1, 31, 1, 1, 1, 7), Counter64()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ifHCInUcastPkts.setStatus('current')
ifHCInMulticastPkts = MibTableColumn((1, 3, 6, 1, 2, 1, 31, 1, 1, 1, 8), Counter64()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ifHCInMulticastPkts.setStatus('current')
ifHCInBroadcastPkts = MibTableColumn((1, 3, 6, 1, 2, 1, 31, 1, 1, 1, 9), Counter64()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ifHCInBroadcastPkts.setStatus('current')
ifHCOutOctets = MibTableColumn((1, 3, 6, 1, 2, 1, 31, 1, 1, 1, 10), Counter64()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ifHCOutOctets.setStatus('current')
ifHCOutUcastPkts = MibTableColumn((1, 3, 6, 1, 2, 1, 31, 1, 1, 1, 11), Counter64()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ifHCOutUcastPkts.setStatus('current')
ifHCOutMulticastPkts = MibTableColumn((1, 3, 6, 1, 2, 1, 31, 1, 1, 1, 12), Counter64()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ifHCOutMulticastPkts.setStatus('current')
ifHCOutBroadcastPkts = MibTableColumn((1, 3, 6, 1, 2, 1, 31, 1, 1, 1, 13), Counter64()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ifHCOutBroadcastPkts.setStatus('current')
ifLinkUpDownTrapEnable = MibTableColumn((1, 3, 6, 1, 2, 1, 31, 1, 1, 1, 14), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(1, 2))).clone(namedValues=NamedValues(("enabled", 1), ("disabled", 2)))).setMaxAccess("readwrite")
if mibBuilder.loadTexts: ifLinkUpDownTrapEnable.setStatus('current')
ifHighSpeed = MibTableColumn((1, 3, 6, 1, 2, 1, 31, 1, 1, 1, 15), Gauge32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ifHighSpeed.setStatus('current')
ifPromiscuousMode = MibTableColumn((1, 3, 6, 1, 2, 1, 31, 1, 1, 1, 16), TruthValue()).setMaxAccess("readwrite")
if mibBuilder.loadTexts: ifPromiscuousMode.setStatus('current')
ifConnectorPresent = MibTableColumn((1, 3, 6, 1, 2, 1, 31, 1, 1, 1, 17), TruthValue()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ifConnectorPresent.setStatus('current')
ifAlias = MibTableColumn((1, 3, 6, 1, 2, 1, 31, 1, 1, 1, 18), DisplayString().subtype(subtypeSpec=ValueSizeConstraint(0, 64))).setMaxAccess("readwrite")
if mibBuilder.loadTexts: ifAlias.setStatus('current')
ifCounterDiscontinuityTime = MibTableColumn((1, 3, 6, 1, 2, 1, 31, 1, 1, 1, 19), TimeStamp()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ifCounterDiscontinuityTime.setStatus('current')
ifStackTable = MibTable((1, 3, 6, 1, 2, 1, 31, 1, 2), )
if mibBuilder.loadTexts: ifStackTable.setStatus('current')
ifStackEntry = MibTableRow((1, 3, 6, 1, 2, 1, 31, 1, 2, 1), ).setIndexNames((0, "IF-MIB", "ifStackHigherLayer"), (0, "IF-MIB", "ifStackLowerLayer"))
if mibBuilder.loadTexts: ifStackEntry.setStatus('current')
ifStackHigherLayer = MibTableColumn((1, 3, 6, 1, 2, 1, 31, 1, 2, 1, 1), InterfaceIndexOrZero())
if mibBuilder.loadTexts: ifStackHigherLayer.setStatus('current')
ifStackLowerLayer = MibTableColumn((1, 3, 6, 1, 2, 1, 31, 1, 2, 1, 2), InterfaceIndexOrZero())
if mibBuilder.loadTexts: ifStackLowerLayer.setStatus('current')
ifStackStatus = MibTableColumn((1, 3, 6, 1, 2, 1, 31, 1, 2, 1, 3), RowStatus()).setMaxAccess("readcreate")
if mibBuilder.loadTexts: ifStackStatus.setStatus('current')
ifStackLastChange = MibScalar((1, 3, 6, 1, 2, 1, 31, 1, 6), TimeTicks()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ifStackLastChange.setStatus('current')
ifRcvAddressTable = MibTable((1, 3, 6, 1, 2, 1, 31, 1, 4), )
if mibBuilder.loadTexts: ifRcvAddressTable.setStatus('current')
ifRcvAddressEntry = MibTableRow((1, 3, 6, 1, 2, 1, 31, 1, 4, 1), ).setIndexNames((0, "IF-MIB", "ifIndex"), (0, "IF-MIB", "ifRcvAddressAddress"))
if mibBuilder.loadTexts: ifRcvAddressEntry.setStatus('current')
ifRcvAddressAddress = MibTableColumn((1, 3, 6, 1, 2, 1, 31, 1, 4, 1, 1), PhysAddress())
if mibBuilder.loadTexts: ifRcvAddressAddress.setStatus('current')
ifRcvAddressStatus = MibTableColumn((1, 3, 6, 1, 2, 1, 31, 1, 4, 1, 2), RowStatus()).setMaxAccess("readcreate")
if mibBuilder.loadTexts: ifRcvAddressStatus.setStatus('current')
ifRcvAddressType = MibTableColumn((1, 3, 6, 1, 2, 1, 31, 1, 4, 1, 3), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(1, 2, 3))).clone(namedValues=NamedValues(("other", 1), ("volatile", 2), ("nonVolatile", 3))).clone('volatile')).setMaxAccess("readcreate")
if mibBuilder.loadTexts: ifRcvAddressType.setStatus('current')
linkDown = NotificationType((1, 3, 6, 1, 6, 3, 1, 1, 5, 3)).setObjects(("IF-MIB", "ifIndex"), ("IF-MIB", "ifAdminStatus"), ("IF-MIB", "ifOperStatus"))
if mibBuilder.loadTexts: linkDown.setStatus('current')
linkUp = NotificationType((1, 3, 6, 1, 6, 3, 1, 1, 5, 4)).setObjects(("IF-MIB", "ifIndex"), ("IF-MIB", "ifAdminStatus"), ("IF-MIB", "ifOperStatus"))
if mibBuilder.loadTexts: linkUp.setStatus('current')
ifConformance = MibIdentifier((1, 3, 6, 1, 2, 1, 31, 2))
ifGroups = MibIdentifier((1, 3, 6, 1, 2, 1, 31, 2, 1))
ifCompliances = MibIdentifier((1, 3, 6, 1, 2, 1, 31, 2, 2))
ifCompliance3 = ModuleCompliance((1, 3, 6, 1, 2, 1, 31, 2, 2, 3)).setObjects(("IF-MIB", "ifGeneralInformationGroup"), ("IF-MIB", "linkUpDownNotificationsGroup"), ("IF-MIB", "ifFixedLengthGroup"), ("IF-MIB", "ifHCFixedLengthGroup"), ("IF-MIB", "ifPacketGroup"), ("IF-MIB", "ifHCPacketGroup"), ("IF-MIB", "ifVHCPacketGroup"), ("IF-MIB", "ifCounterDiscontinuityGroup"), ("IF-MIB", "ifRcvAddressGroup"))

if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    ifCompliance3 = ifCompliance3.setStatus('current')
ifGeneralInformationGroup = ObjectGroup((1, 3, 6, 1, 2, 1, 31, 2, 1, 10)).setObjects(("IF-MIB", "ifIndex"), ("IF-MIB", "ifDescr"), ("IF-MIB", "ifType"), ("IF-MIB", "ifSpeed"), ("IF-MIB", "ifPhysAddress"), ("IF-MIB", "ifAdminStatus"), ("IF-MIB", "ifOperStatus"), ("IF-MIB", "ifLastChange"), ("IF-MIB", "ifLinkUpDownTrapEnable"), ("IF-MIB", "ifConnectorPresent"), ("IF-MIB", "ifHighSpeed"), ("IF-MIB", "ifName"), ("IF-MIB", "ifNumber"), ("IF-MIB", "ifAlias"), ("IF-MIB", "ifTableLastChange"))
if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    ifGeneralInformationGroup = ifGeneralInformationGroup.setStatus('current')
ifFixedLengthGroup = ObjectGroup((1, 3, 6, 1, 2, 1, 31, 2, 1, 2)).setObjects(("IF-MIB", "ifInOctets"), ("IF-MIB", "ifOutOctets"), ("IF-MIB", "ifInUnknownProtos"), ("IF-MIB", "ifInErrors"), ("IF-MIB", "ifOutErrors"))
if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    ifFixedLengthGroup = ifFixedLengthGroup.setStatus('current')
ifHCFixedLengthGroup = ObjectGroup((1, 3, 6, 1, 2, 1, 31, 2, 1, 3)).setObjects(("IF-MIB", "ifHCInOctets"), ("IF-MIB", "ifHCOutOctets"), ("IF-MIB", "ifInOctets"), ("IF-MIB", "ifOutOctets"), ("IF-MIB", "ifInUnknownProtos"), ("IF-MIB", "ifInErrors"), ("IF-MIB", "ifOutErrors"))
if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    ifHCFixedLengthGroup = ifHCFixedLengthGroup.setStatus('current')
ifPacketGroup = ObjectGroup((1, 3, 6, 1, 2, 1, 31, 2, 1, 4)).setObjects(("IF-MIB", "ifInOctets"), ("IF-MIB", "ifOutOctets"), ("IF-MIB", "ifInUnknownProtos"), ("IF-MIB", "ifInErrors"), ("IF-MIB", "ifOutErrors"), ("IF-MIB", "ifMtu"), ("IF-MIB", "ifInUcastPkts"), ("IF-MIB", "ifInMulticastPkts"), ("IF-MIB", "ifInBroadcastPkts"), ("IF-MIB", "ifInDiscards"), ("IF-MIB", "ifOutUcastPkts"), ("IF-MIB", "ifOutMulticastPkts"), ("IF-MIB", "ifOutBroadcastPkts"), ("IF-MIB", "ifOutDiscards"), ("IF-MIB", "ifPromiscuousMode"))
if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    ifPacketGroup = ifPacketGroup.setStatus('current')
ifHCPacketGroup = ObjectGroup((1, 3, 6, 1, 2, 1, 31, 2, 1, 5)).setObjects(("IF-MIB", "ifHCInOctets"), ("IF-MIB", "ifHCOutOctets"), ("IF-MIB", "ifInOctets"), ("IF-MIB", "ifOutOctets"), ("IF-MIB", "ifInUnknownProtos"), ("IF-MIB", "ifInErrors"), ("IF-MIB", "ifOutErrors"), ("IF-MIB", "ifMtu"), ("IF-MIB", "ifInUcastPkts"), ("IF-MIB", "ifInMulticastPkts"), ("IF-MIB", "ifInBroadcastPkts"), ("IF-MIB", "ifInDiscards"), ("IF-MIB", "ifOutUcastPkts"), ("IF-MIB", "ifOutMulticastPkts"), ("IF-MIB", "ifOutBroadcastPkts"), ("IF-MIB", "ifOutDiscards"), ("IF-MIB", "ifPromiscuousMode"))
if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    ifHCPacketGroup = ifHCPacketGroup.setStatus('current')
ifVHCPacketGroup = ObjectGroup((1, 3, 6, 1, 2, 1, 31, 2, 1, 6)).setObjects(("IF-MIB", "ifHCInUcastPkts"), ("IF-MIB", "ifHCInMulticastPkts"), ("IF-MIB", "ifHCInBroadcastPkts"), ("IF-MIB", "ifHCOutUcastPkts"), ("IF-MIB", "ifHCOutMulticastPkts"), ("IF-MIB", "ifHCOutBroadcastPkts"), ("IF-MIB", "ifHCInOctets"), ("IF-MIB", "ifHCOutOctets"), ("IF-MIB", "ifInOctets"), ("IF-MIB", "ifOutOctets"), ("IF-MIB", "ifInUnknownProtos"), ("IF-MIB", "ifInErrors"), ("IF-MIB", "ifOutErrors"), ("IF-MIB", "ifMtu"), ("IF-MIB", "ifInUcastPkts"), ("IF-MIB", "ifInMulticastPkts"), ("IF-MIB", "ifInBroadcastPkts"), ("IF-MIB", "ifInDiscards"), ("IF-MIB", "ifOutUcastPkts"), ("IF-MIB", "ifOutMulticastPkts"), ("IF-MIB", "ifOutBroadcastPkts"), ("IF-MIB", "ifOutDiscards"), ("IF-MIB", "ifPromiscuousMode"))
if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    ifVHCPacketGroup = ifVHCPacketGroup.setStatus('current')
ifRcvAddressGroup = ObjectGroup((1, 3, 6, 1, 2, 1, 31, 2, 1, 7)).setObjects(("IF-MIB", "ifRcvAddressStatus"), ("IF-MIB", "ifRcvAddressType"))
if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    ifRcvAddressGroup = ifRcvAddressGroup.setStatus('current')
ifStackGroup2 = ObjectGroup((1, 3, 6, 1, 2, 1, 31, 2, 1, 11)).setObjects(("IF-MIB", "ifStackStatus"), ("IF-MIB", "ifStackLastChange"))
if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    ifStackGroup2 = ifStackGroup2.setStatus('current')
ifCounterDiscontinuityGroup = ObjectGroup((1, 3, 6, 1, 2, 1, 31, 2, 1, 13)).setObjects(("IF-MIB", "ifCounterDiscontinuityTime"))
if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    ifCounterDiscontinuityGroup = ifCounterDiscontinuityGroup.setStatus('current')
linkUpDownNotificationsGroup = NotificationGroup((1, 3, 6, 1, 2, 1, 31, 2, 1, 14)).setObjects(("IF-MIB", "linkUp"), ("IF-MIB", "linkDown"))
if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    linkUpDownNotificationsGroup = linkUpDownNotificationsGroup.setStatus('current')
ifTestTable = MibTable((1, 3, 6, 1, 2, 1, 31, 1, 3), )
if mibBuilder.loadTexts: ifTestTable.setStatus('deprecated')
ifTestEntry = MibTableRow((1, 3, 6, 1, 2, 1, 31, 1, 3, 1), )
ifEntry.registerAugmentions(("IF-MIB", "ifTestEntry"))
ifTestEntry.setIndexNames(*ifEntry.getIndexNames())
if mibBuilder.loadTexts: ifTestEntry.setStatus('deprecated')
ifTestId = MibTableColumn((1, 3, 6, 1, 2, 1, 31, 1, 3, 1, 1), TestAndIncr()).setMaxAccess("readwrite")
if mibBuilder.loadTexts: ifTestId.setStatus('deprecated')
ifTestStatus = MibTableColumn((1, 3, 6, 1, 2, 1, 31, 1, 3, 1, 2), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(1, 2))).clone(namedValues=NamedValues(("notInUse", 1), ("inUse", 2)))).setMaxAccess("readwrite")
if mibBuilder.loadTexts: ifTestStatus.setStatus('deprecated')
ifTestType = MibTableColumn((1, 3, 6, 1, 2, 1, 31, 1, 3, 1, 3), AutonomousType()).setMaxAccess("readwrite")
if mibBuilder.loadTexts: ifTestType.setStatus('deprecated')
ifTestResult = MibTableColumn((1, 3, 6, 1, 2, 1, 31, 1, 3, 1, 4), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(1, 2, 3, 4, 5, 6, 7))).clone(namedValues=NamedValues(("none", 1), ("success", 2), ("inProgress", 3), ("notSupported", 4), ("unAbleToRun", 5), ("aborted", 6), ("failed", 7)))).setMaxAccess("readonly")
if mibBuilder.loadTexts: ifTestResult.setStatus('deprecated')
ifTestCode = MibTableColumn((1, 3, 6, 1, 2, 1, 31, 1, 3, 1, 5), ObjectIdentifier()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ifTestCode.setStatus('deprecated')
ifTestOwner = MibTableColumn((1, 3, 6, 1, 2, 1, 31, 1, 3, 1, 6), OwnerString()).setMaxAccess("readwrite")
if mibBuilder.loadTexts: ifTestOwner.setStatus('deprecated')
ifGeneralGroup = ObjectGroup((1, 3, 6, 1, 2, 1, 31, 2, 1, 1)).setObjects(("IF-MIB", "ifDescr"), ("IF-MIB", "ifType"), ("IF-MIB", "ifSpeed"), ("IF-MIB", "ifPhysAddress"), ("IF-MIB", "ifAdminStatus"), ("IF-MIB", "ifOperStatus"), ("IF-MIB", "ifLastChange"), ("IF-MIB", "ifLinkUpDownTrapEnable"), ("IF-MIB", "ifConnectorPresent"), ("IF-MIB", "ifHighSpeed"), ("IF-MIB", "ifName"))
if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    ifGeneralGroup = ifGeneralGroup.setStatus('deprecated')
ifTestGroup = ObjectGroup((1, 3, 6, 1, 2, 1, 31, 2, 1, 8)).setObjects(("IF-MIB", "ifTestId"), ("IF-MIB", "ifTestStatus"), ("IF-MIB", "ifTestType"), ("IF-MIB", "ifTestResult"), ("IF-MIB", "ifTestCode"), ("IF-MIB", "ifTestOwner"))
if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    ifTestGroup = ifTestGroup.setStatus('deprecated')
ifStackGroup = ObjectGroup((1, 3, 6, 1, 2, 1, 31, 2, 1, 9)).setObjects(("IF-MIB", "ifStackStatus"))
if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    ifStackGroup = ifStackGroup.setStatus('deprecated')
ifOldObjectsGroup = ObjectGroup((1, 3, 6, 1, 2, 1, 31, 2, 1, 12)).setObjects(("IF-MIB", "ifInNUcastPkts"), ("IF-MIB", "ifOutNUcastPkts"), ("IF-MIB", "ifOutQLen"), ("IF-MIB", "ifSpecific"))
if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    ifOldObjectsGroup = ifOldObjectsGroup.setStatus('deprecated')
ifCompliance = ModuleCompliance((1, 3, 6, 1, 2, 1, 31, 2, 2, 1)).setObjects(("IF-MIB", "ifGeneralGroup"), ("IF-MIB", "ifStackGroup"), ("IF-MIB", "ifFixedLengthGroup"), ("IF-MIB", "ifHCFixedLengthGroup"), ("IF-MIB", "ifPacketGroup"), ("IF-MIB", "ifHCPacketGroup"), ("IF-MIB", "ifTestGroup"), ("IF-MIB", "ifRcvAddressGroup"))

if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    ifCompliance = ifCompliance.setStatus('deprecated')
ifCompliance2 = ModuleCompliance((1, 3, 6, 1, 2, 1, 31, 2, 2, 2)).setObjects(("IF-MIB", "ifGeneralInformationGroup"), ("IF-MIB", "ifStackGroup2"), ("IF-MIB", "ifCounterDiscontinuityGroup"), ("IF-MIB", "ifFixedLengthGroup"), ("IF-MIB", "ifHCFixedLengthGroup"), ("IF-MIB", "ifPacketGroup"), ("IF-MIB", "ifHCPacketGroup"), ("IF-MIB", "ifRcvAddressGroup"))

if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    ifCompliance2 = ifCompliance2.setStatus('deprecated')
mibBuilder.exportSymbols("IF-MIB", ifInBroadcastPkts=ifInBroadcastPkts, ifRcvAddressStatus=ifRcvAddressStatus, ifHCOutMulticastPkts=ifHCOutMulticastPkts, ifTableLastChange=ifTableLastChange, ifXEntry=ifXEntry, ifNumber=ifNumber, ifLinkUpDownTrapEnable=ifLinkUpDownTrapEnable, ifMIBObjects=ifMIBObjects, ifGeneralGroup=ifGeneralGroup, ifHCOutBroadcastPkts=ifHCOutBroadcastPkts, ifCompliance=ifCompliance, ifConnectorPresent=ifConnectorPresent, ifOutErrors=ifOutErrors, ifStackGroup2=ifStackGroup2, ifStackGroup=ifStackGroup, ifInMulticastPkts=ifInMulticastPkts, ifXTable=ifXTable, InterfaceIndexOrZero=InterfaceIndexOrZero, ifRcvAddressTable=ifRcvAddressTable, ifTestId=ifTestId, ifPhysAddress=ifPhysAddress, ifTestOwner=ifTestOwner, ifRcvAddressType=ifRcvAddressType, ifOldObjectsGroup=ifOldObjectsGroup, ifInNUcastPkts=ifInNUcastPkts, ifOutMulticastPkts=ifOutMulticastPkts, ifCounterDiscontinuityTime=ifCounterDiscontinuityTime, ifCompliances=ifCompliances, ifPacketGroup=ifPacketGroup, ifAdminStatus=ifAdminStatus, ifTestCode=ifTestCode, ifInOctets=ifInOctets, ifIndex=ifIndex, ifOutQLen=ifOutQLen, ifInUcastPkts=ifInUcastPkts, ifCompliance3=ifCompliance3, ifTestTable=ifTestTable, ifInDiscards=ifInDiscards, ifEntry=ifEntry, ifHCInOctets=ifHCInOctets, ifStackLastChange=ifStackLastChange, ifAlias=ifAlias, linkDown=linkDown, ifName=ifName, linkUpDownNotificationsGroup=linkUpDownNotificationsGroup, ifRcvAddressAddress=ifRcvAddressAddress, ifStackLowerLayer=ifStackLowerLayer, ifStackHigherLayer=ifStackHigherLayer, ifMIB=ifMIB, ifDescr=ifDescr, ifHCOutOctets=ifHCOutOctets, ifStackEntry=ifStackEntry, ifOutDiscards=ifOutDiscards, ifOutOctets=ifOutOctets, linkUp=linkUp, ifFixedLengthGroup=ifFixedLengthGroup, ifHighSpeed=ifHighSpeed, InterfaceIndex=InterfaceIndex, ifCompliance2=ifCompliance2, ifHCInUcastPkts=ifHCInUcastPkts, ifGeneralInformationGroup=ifGeneralInformationGroup, ifTestEntry=ifTestEntry, ifInUnknownProtos=ifInUnknownProtos, ifTable=ifTable, PYSNMP_MODULE_ID=ifMIB, ifTestResult=ifTestResult, ifOutBroadcastPkts=ifOutBroadcastPkts, ifHCInBroadcastPkts=ifHCInBroadcastPkts, OwnerString=OwnerString, ifSpecific=ifSpecific, ifType=ifType, ifOutNUcastPkts=ifOutNUcastPkts, ifRcvAddressGroup=ifRcvAddressGroup, ifHCPacketGroup=ifHCPacketGroup, ifCounterDiscontinuityGroup=ifCounterDiscontinuityGroup, ifHCInMulticastPkts=ifHCInMulticastPkts, interfaces=interfaces, ifInErrors=ifInErrors, ifConformance=ifConformance, ifOutUcastPkts=ifOutUcastPkts, ifStackStatus=ifStackStatus, ifRcvAddressEntry=ifRcvAddressEntry, ifGroups=ifGroups, ifVHCPacketGroup=ifVHCPacketGroup, ifTestStatus=ifTestStatus, ifPromiscuousMode=ifPromiscuousMode, ifLastChange=ifLastChange, ifHCOutUcastPkts=ifHCOutUcastPkts, ifTestGroup=ifTestGroup, ifHCFixedLengthGroup=ifHCFixedLengthGroup, ifStackTable=ifStackTable, ifOperStatus=ifOperStatus, ifMtu=ifMtu, ifSpeed=ifSpeed, ifTestType=ifTestType)



"""Tests for the SNMP agent module."""

import asyncio
import subprocess
import time
from unittest.mock import MagicMock, patch

import pytest

from zino.config.models import PollDevice
from zino.snmp import SNMP
from zino.snmp.agent import ZinoSnmpAgent


class TestZinoSnmpAgent:
    """Test the ZinoSnmpAgent class."""

    def test_when_initialized_without_arguments_then_it_should_use_default_values(self):
        """Test that __init__ sets sensible defaults."""
        agent = ZinoSnmpAgent()
        assert agent.listen_address == "0.0.0.0"
        assert agent.listen_port == 8000
        assert agent.community == "public"
        assert agent.start_time is not None
        assert not agent._running
        assert agent.snmp_engine is None
        assert agent.transport_dispatcher is None

    def test_when_initialized_with_custom_values_then_it_should_store_them(self):
        """Test that __init__ accepts custom configuration."""
        start_time = time.time() - 100
        agent = ZinoSnmpAgent(
            listen_address="127.0.0.1",
            listen_port=8001,
            community="secret",
            start_time=start_time,
        )
        assert agent.listen_address == "127.0.0.1"
        assert agent.listen_port == 8001
        assert agent.community == "secret"
        assert agent.start_time == start_time

    def test_when_get_uptime_called_then_it_should_return_seconds_since_start_time(self):
        """Test that _get_uptime calculates uptime correctly."""
        start_time = time.time() - 42  # Started 42 seconds ago
        agent = ZinoSnmpAgent(start_time=start_time)
        uptime = agent._get_uptime()
        # Allow for small timing differences
        assert 41 <= uptime <= 43

    def test_when_get_uptime_called_then_it_should_return_integer_type(self):
        """Test that _get_uptime returns an integer."""
        agent = ZinoSnmpAgent()
        uptime = agent._get_uptime()
        assert isinstance(uptime, int)

    def test_when_setup_engine_called_then_it_should_initialize_snmp_engine_and_dispatcher(self):
        """Test that _setup_engine initializes the engine and dispatcher attributes."""
        agent = ZinoSnmpAgent(listen_address="127.0.0.1", listen_port=8001, community="public")

        # Mock the entire method to avoid complex SNMP setup
        with patch.object(agent, "_setup_engine") as mock_setup:

            def setup_side_effect():
                agent.snmp_engine = MagicMock()
                agent.transport_dispatcher = MagicMock()

            mock_setup.side_effect = setup_side_effect

            agent._setup_engine()

            # Check that attributes are set
            assert agent.snmp_engine is not None
            assert agent.transport_dispatcher is not None
            mock_setup.assert_called_once()

    @patch("zino.snmp.agent.builder")
    def test_when_register_zino_uptime_called_then_it_should_load_and_export_mib_symbols(self, mock_builder):
        """Test that _register_zino_uptime loads required MIB modules."""
        agent = ZinoSnmpAgent()
        mock_mib_instrum = MagicMock()
        mock_mib_builder = MagicMock()
        mock_mib_instrum.mibBuilder = mock_mib_builder

        # Mock the MIB symbols
        mock_zino_uptime = MagicMock()
        mock_zino_uptime.getName.return_value = (1, 3, 6, 1, 4, 1, 2428, 130, 1, 1, 1)
        mock_zino_uptime.getSyntax.return_value = MagicMock()
        mock_mib_scalar_instance = MagicMock()

        def mock_import_symbols(module, *symbols):
            if module == "ZINO-MIB":
                return (mock_zino_uptime,)
            elif module == "SNMPv2-SMI":
                return (mock_mib_scalar_instance,)
            return ()

        mock_mib_builder.importSymbols = mock_import_symbols

        agent._register_zino_uptime(mock_mib_instrum)

        # Check that MIB modules were loaded
        mock_mib_builder.loadModules.assert_called()

    @pytest.mark.asyncio
    async def test_when_start_called_then_it_should_set_running_flag_to_true(self):
        """Test that start() sets the _running flag and calls setup."""
        agent = ZinoSnmpAgent(listen_port=18000)

        with patch.object(agent, "_setup_engine") as mock_setup:
            mock_dispatcher = MagicMock()
            mock_dispatcher.runDispatcher.return_value = None
            agent.transport_dispatcher = mock_dispatcher

            # Start agent in background and let it run briefly
            task = asyncio.create_task(agent.open())
            await asyncio.sleep(0.01)

            # Check that setup was called and running flag is set
            mock_setup.assert_called_once()
            assert agent._running

            # Clean up
            agent.close()
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                pass

    @pytest.mark.asyncio
    async def test_when_start_fails_during_setup_then_it_should_raise_exception_and_clear_running_flag(self):
        """Test that start() handles setup failures gracefully."""
        agent = ZinoSnmpAgent()

        with patch.object(agent, "_setup_engine", side_effect=Exception("Setup failed")):
            with pytest.raises(Exception, match="Setup failed"):
                await agent.open()

            assert not agent._running

    def test_when_stop_called_on_running_agent_then_it_should_close_dispatchers_and_clear_running_flag(self):
        """Test that stop() clears the _running flag."""
        agent = ZinoSnmpAgent()
        agent._running = True
        agent.transport_dispatcher = MagicMock()
        agent.snmp_engine = MagicMock()

        agent.close()

        assert not agent._running
        agent.transport_dispatcher.closeDispatcher.assert_called_once()
        agent.snmp_engine.transportDispatcher.closeDispatcher.assert_called_once()

    def test_when_stop_called_on_non_running_agent_then_it_should_do_nothing(self):
        """Test that stop() does nothing when agent is not running."""
        agent = ZinoSnmpAgent()
        agent._running = False
        agent.close()  # Should not raise any exceptions
        assert not agent._running

    def test_when_stop_called_with_none_dispatchers_then_it_should_not_raise_exception(self):
        """Test that stop() handles None dispatcher gracefully."""
        agent = ZinoSnmpAgent()
        agent._running = True
        agent.transport_dispatcher = None
        agent.snmp_engine = None

        agent.close()  # Should not raise exceptions
        assert not agent._running


class TestSetupEngineConfiguration:
    """Test different configuration branches in _setup_engine method."""

    @patch("zino.snmp.agent.cmdrsp")
    @patch("zino.snmp.agent.context")
    @patch("zino.snmp.agent.builder")
    @patch("zino.snmp.agent.config")
    @patch("zino.snmp.agent.AsyncioDispatcher")
    @patch("zino.snmp.agent.engine")
    @patch("zino.snmp.agent.udp6")
    @patch("zino.snmp.agent.udp")
    def test_when_address_is_ipv6_then_it_should_use_ipv6_transport(
        self, mock_udp, mock_udp6, mock_engine, mock_dispatcher, mock_config, mock_builder, mock_context, mock_cmdrsp
    ):
        """Test that IPv6 addresses result in IPv6 transport configuration."""
        agent = ZinoSnmpAgent(listen_address="::1", listen_port=8001)

        mock_transport = MagicMock()
        mock_udp6.Udp6AsyncioTransport.return_value.openServerMode.return_value = mock_transport
        mock_snmp_engine = MagicMock()
        mock_engine.SnmpEngine.return_value = mock_snmp_engine

        with patch.object(agent, "_register_zino_uptime"):
            agent._setup_engine()

        # Verify IPv6 transport was used
        mock_udp6.Udp6AsyncioTransport.assert_called_once()
        mock_udp.UdpAsyncioTransport.assert_not_called()

    @patch("zino.snmp.agent.cmdrsp")
    @patch("zino.snmp.agent.context")
    @patch("zino.snmp.agent.builder")
    @patch("zino.snmp.agent.config")
    @patch("zino.snmp.agent.AsyncioDispatcher")
    @patch("zino.snmp.agent.engine")
    @patch("zino.snmp.agent.udp6")
    @patch("zino.snmp.agent.udp")
    def test_when_address_is_ipv4_then_it_should_use_ipv4_transport(
        self, mock_udp, mock_udp6, mock_engine, mock_dispatcher, mock_config, mock_builder, mock_context, mock_cmdrsp
    ):
        """Test that IPv4 addresses result in IPv4 transport configuration."""
        agent = ZinoSnmpAgent(listen_address="192.168.1.1", listen_port=8001)

        mock_transport = MagicMock()
        mock_udp.UdpAsyncioTransport.return_value.openServerMode.return_value = mock_transport
        mock_snmp_engine = MagicMock()
        mock_engine.SnmpEngine.return_value = mock_snmp_engine

        with patch.object(agent, "_register_zino_uptime"):
            agent._setup_engine()

        # Verify IPv4 transport was used
        mock_udp.UdpAsyncioTransport.assert_called_once()
        mock_udp6.Udp6AsyncioTransport.assert_not_called()

    @patch("zino.snmp.agent.cmdrsp")
    @patch("zino.snmp.agent.context")
    @patch("zino.snmp.agent.builder")
    @patch("zino.snmp.agent.config")
    @patch("zino.snmp.agent.AsyncioDispatcher")
    @patch("zino.snmp.agent.engine")
    @patch("zino.snmp.agent.udp")
    def test_when_address_is_hostname_then_it_should_use_ipv4_transport(
        self, mock_udp, mock_engine, mock_dispatcher, mock_config, mock_builder, mock_context, mock_cmdrsp
    ):
        """Test that hostnames default to IPv4 transport."""
        agent = ZinoSnmpAgent(listen_address="localhost", listen_port=8001)

        mock_transport = MagicMock()
        mock_udp.UdpAsyncioTransport.return_value.openServerMode.return_value = mock_transport
        mock_snmp_engine = MagicMock()
        mock_engine.SnmpEngine.return_value = mock_snmp_engine

        with patch.object(agent, "_register_zino_uptime"):
            agent._setup_engine()

        # Verify IPv4 transport was used for hostname
        mock_udp.UdpAsyncioTransport.assert_called_once()

    @patch("zino.snmp.agent.cmdrsp")
    @patch("zino.snmp.agent.context")
    @patch("zino.snmp.agent.builder")
    @patch("zino.snmp.agent.config")
    @patch("zino.snmp.agent.AsyncioDispatcher")
    @patch("zino.snmp.agent.engine")
    @patch("zino.snmp.agent.udp")
    def test_when_specific_community_then_it_should_only_configure_that_community(
        self, mock_udp, mock_engine, mock_dispatcher, mock_config, mock_builder, mock_context, mock_cmdrsp
    ):
        """Test that specific community configuration is applied."""
        agent = ZinoSnmpAgent(community="test-community")

        mock_transport = MagicMock()
        mock_udp.UdpAsyncioTransport.return_value.openServerMode.return_value = mock_transport
        mock_snmp_engine = MagicMock()
        mock_engine.SnmpEngine.return_value = mock_snmp_engine

        with patch.object(agent, "_register_zino_uptime"):
            agent._setup_engine()

        # Verify specific community was configured
        calls = mock_config.addV1System.call_args_list
        assert len([c for c in calls if "test-community" in c[0]]) == 1
        # Should not have added default community value
        assert len([c for c in calls if "public" in c[0]]) == 0

    @patch("zino.snmp.agent.cmdrsp")
    @patch("zino.snmp.agent.context")
    @patch("zino.snmp.agent.builder")
    @patch("zino.snmp.agent.config")
    @patch("zino.snmp.agent.AsyncioDispatcher")
    @patch("zino.snmp.agent.engine")
    @patch("zino.snmp.agent.udp")
    def test_when_no_community_then_it_should_configure_public_community(
        self, mock_udp, mock_engine, mock_dispatcher, mock_config, mock_builder, mock_context, mock_cmdrsp
    ):
        """Test that no community means multiple communities are accepted."""
        agent = ZinoSnmpAgent(community=None)

        mock_transport = MagicMock()
        mock_udp.UdpAsyncioTransport.return_value.openServerMode.return_value = mock_transport
        mock_snmp_engine = MagicMock()
        mock_engine.SnmpEngine.return_value = mock_snmp_engine

        with patch.object(agent, "_register_zino_uptime"):
            agent._setup_engine()

        calls = mock_config.addV1System.call_args_list
        assert any("public" in str(c) for c in calls)


@pytest.mark.asyncio
async def test_when_agent_is_queried_then_it_should_return_uptime(running_agent):
    """Integration test that the agent responds to SNMP queries using Zino's SNMP backend."""
    port = running_agent

    # Create a PollDevice pointing to our agent
    device = PollDevice(name="test-agent", address="127.0.0.1", port=port, community="public")

    # Use Zino's SNMP backend to query the agent
    with SNMP(device) as snmp_session:
        # Query ZINO-MIB::zinoUpTime.0
        response = await snmp_session.get("ZINO-MIB", "zinoUpTime", 0)

        # Verify we got a response
        assert response is not None
        assert response.value is not None

        # Check that uptime is reasonable (0-10 seconds for a just-started agent)
        uptime = int(response.value)
        assert 0 <= uptime <= 10, f"Unexpected uptime value: {uptime}"


@pytest.mark.asyncio
async def test_when_agent_runs_then_uptime_should_increase(running_agent):
    """Test that zinoUpTime value increases over time."""
    port = running_agent

    device = PollDevice(name="test-agent", address="127.0.0.1", port=port, community="public")

    with SNMP(device) as snmp_session:
        # First query
        response1 = await snmp_session.get("ZINO-MIB", "zinoUpTime", 0)
        uptime1 = int(response1.value)

        # Wait 2 seconds
        await asyncio.sleep(2)

        # Second query
        response2 = await snmp_session.get("ZINO-MIB", "zinoUpTime", 0)
        uptime2 = int(response2.value)

        # Uptime should have increased by approximately 2 seconds
        diff = uptime2 - uptime1
        assert 1 <= diff <= 3, f"Uptime difference was {diff}, expected ~2"


@pytest.fixture
async def running_agent(unused_udp_port):
    """Start the SNMP agent as a subprocess and ensure it's ready."""
    process = subprocess.Popen(
        ["python", "-m", "zino.snmp.agent", "--port", str(unused_udp_port)],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )

    # Give the agent time to start up
    await asyncio.sleep(1)

    try:
        yield unused_udp_port
    finally:
        process.terminate()
        try:
            process.wait(timeout=5)
        except subprocess.TimeoutExpired:
            process.kill()
            process.wait()



import pytest

from zino.state import ZinoState
from zino.statemodels import DeviceState, Port, PortStateEvent, ReachabilityEvent


@pytest.fixture
def state() -> ZinoState:
    return ZinoState()


@pytest.fixture
def port() -> Port:
    return Port(ifindex=1, ifdescr="port", ifalias="portalias")


@pytest.fixture
def device(state, port) -> DeviceState:
    device = state.devices.get("device")
    device.ports[port.ifindex] = port
    return device


@pytest.fixture
def reachability_event(state, device) -> ReachabilityEvent:
    event = state.events.create_event(device.name, None, ReachabilityEvent)
    state.events.commit(event)
    return state.events.get(event.router, event.subindex, ReachabilityEvent)


@pytest.fixture
def portstate_event(state, device, port) -> PortStateEvent:
    event = state.events.create_event(device.name, port.ifindex, PortStateEvent)
    event.ifindex = port.ifindex
    state.events.commit(event)
    return state.events.get(event.router, event.subindex, PortStateEvent)



"""Zino SNMP back-ends"""

import importlib
import logging
import os.path
from types import ModuleType
from typing import TYPE_CHECKING
from weakref import WeakValueDictionary

from zino.config.models import PollDevice
from zino.snmp.base import SNMPBackendNotLoaded

if TYPE_CHECKING:
    from zino.snmp.pysnmp_backend import SNMP


_logger = logging.getLogger(__name__)
_snmp_sessions = WeakValueDictionary()
_selected_backend = None
_BACKEND_MAP = {"pysnmp": "pysnmp_backend", "netsnmp": "netsnmpy_backend", None: "netsnmpy_backend"}


def import_snmp_backend(backend: str = None) -> ModuleType:
    """Import and return the selected SNMP back-end. The first time this is called,
    the back-end is selectable, but defaults to "netsnmp".  Subsequent calls that omit backend will
    return the back-end that was selected by the first call.
    """
    global _selected_backend, _snmp

    if backend is None and _selected_backend is not None:
        return _selected_backend

    module_name = _BACKEND_MAP.get(backend)
    if not module_name:
        raise ValueError(f"Unknown SNMP backend: {backend}")

    _logger.info("Loading SNMP backend: %s (%s)", backend, module_name)
    module = importlib.import_module(f"zino.snmp.{module_name}")
    if _selected_backend is None:
        _selected_backend = module
        # Hack to import all public symbols from the module into the current namespace
        globals().update({name: getattr(module, name) for name in dir(module) if not name.startswith("_")})
    return module


def get_snmp_session(device: PollDevice) -> "SNMP":
    """Create or re-use an existing SNMP session for a device.

    This keeps a registry of existing/re-usable SNMP sessions so we avoid duplicate instances and over-use of sockets
    """
    global _selected_backend
    if _selected_backend is None:
        raise SNMPBackendNotLoaded()

    if _snmp_sessions is None:  # Tests suite may disable session re-use
        return _selected_backend.SNMP(device)
    # We generate a session key based on every attribute that affects how the SNMP session is set up:
    key = (device.address, device.community, device.snmpversion)
    session = _snmp_sessions.get(key)
    if not session:
        session = _snmp_sessions[key] = _selected_backend.SNMP(device)
    return session


def get_vendored_mib_directory():
    """Returns the path to the vendored MIB directory"""
    return os.path.join(os.path.dirname(__file__), "mibs")



import io

import pytest

from zino.config.models import PollDevice
from zino.config.polldevs import (
    InvalidConfiguration,
    _contains_defaults,
    _parse_defaults,
    _read_conf_sections,
    read_polldevs,
)


class TestReadPolldevs:
    def test_should_generate_two_polldevices_from_test_config(self, polldevs_conf):
        result, _ = read_polldevs(polldevs_conf)
        assert len(result) == 2
        assert all(isinstance(device, PollDevice) for device in result.values())

    def test_should_return_default_values_from_test_config(self, polldevs_conf):
        _, defaults = read_polldevs(polldevs_conf)
        assert "community" in defaults
        assert defaults["community"] == "foobar"
        assert "domain" in defaults
        assert defaults["domain"] == "uninett.no"

    def test_should_use_default_values_in_polldevices_generated_from_test_config(self, polldevs_conf):
        result, _ = read_polldevs(polldevs_conf)
        assert all(device.community == "foobar" for device in result.values())
        assert all(device.domain == "uninett.no" for device in result.values())

    def test_when_hcounters_is_in_config_it_should_parse_correctly(self, polldevs_conf_with_hcounters):
        result, _ = read_polldevs(polldevs_conf_with_hcounters)
        assert len(result) == 2
        assert all(device.hcounters for device in result.values())


class TestReadInvalidPolldevs:
    def test_should_raise_exception(self, invalid_polldevs_conf):
        with pytest.raises(InvalidConfiguration):
            read_polldevs(invalid_polldevs_conf)

    def test_should_have_filename_in_exception(self, invalid_polldevs_conf):
        with pytest.raises(InvalidConfiguration) as e:
            read_polldevs(invalid_polldevs_conf)
        assert "polldevs.cf" in str(e.value)

    def test_should_have_line_number_in_exception(self, invalid_polldevs_conf):
        with pytest.raises(InvalidConfiguration) as e:
            read_polldevs(invalid_polldevs_conf)
        assert "2" in str(e.value)

    def test_exception_should_include_device_name_on_missing_address(self, missing_device_address_polldevs_conf):
        with pytest.raises(InvalidConfiguration) as e:
            read_polldevs(missing_device_address_polldevs_conf)
        assert "example-gw" in str(e.value)

    def test_exception_should_include_missing_attribute_on_missing_address(self, missing_device_address_polldevs_conf):
        with pytest.raises(InvalidConfiguration) as e:
            read_polldevs(missing_device_address_polldevs_conf)
        assert "Field required ('address')" in str(e.value)


class TestReadConfSections:
    def test_when_file_is_empty_it_should_return_nothing(self):
        data = io.StringIO("")
        result = list(_read_conf_sections(data))
        assert not result

    def test_when_file_contains_two_sections_it_should_yield_two_dicts(self):
        data = io.StringIO(
            """
            name: zaphod

            name: ford
            """
        )
        result = list(_read_conf_sections(data))
        assert len(result) == 2
        assert all(isinstance(block, dict) for lineno, block in result)

    def test_when_file_contains_comments_they_should_be_ignored(self):
        data = io.StringIO(
            """
            # comment: just a comment
            name: zaphod
            #address: 192.168.0.1
            address: 127.0.0.1
            """
        )
        expected = {"name": "zaphod", "address": "127.0.0.1"}
        result = list(block for lineno, block in _read_conf_sections(data))
        assert result == [expected]

    def test_when_file_contains_non_assignments_it_should_fail(self):
        data = io.StringIO(
            """
            random data
            not a config file
            """
        )
        with pytest.raises(InvalidConfiguration):
            list(_read_conf_sections(data))


class TestContainsDefaults:
    def test_when_section_contains_at_least_one_default_key_it_should_return_true(self):
        section = {"default value": "foobar", "other": "cromulent"}
        assert _contains_defaults(section)

    def test_when_section_contains_no_default_keys_it_should_return_false(self):
        section = {"value 1": "foobar", "value 2": "cromulent"}
        assert not _contains_defaults(section)


class TestParseDefaults:
    def test_all_default_values_should_be_returned(self):
        section = {"default value1": "foobar", "default value2": "cromulent", "value3": "zaphod"}
        expected = {"value1": "foobar", "value2": "cromulent"}
        assert _parse_defaults(section) == expected


@pytest.fixture
def missing_device_address_polldevs_conf(tmp_path):
    name = tmp_path.joinpath("polldevs.cf")
    with open(name, "w") as conf:
        conf.write(
            """# polldevs test config
            default interval: 5
            default community: foobar
            default domain: uninett.no
            default statistics: yes
            default snmpversion: v2c

            name: example-gw
            """
        )
    yield name


@pytest.fixture
def polldevs_conf_with_hcounters(tmp_path):
    name = tmp_path.joinpath("polldevs.cf")
    with open(name, "w") as conf:
        conf.write(
            """# polldevs test config
            default interval: 5
            default community: foobar
            default domain: uninett.no
            default statistics: yes
            default snmpversion: v2c
            default hcounters: yes

            name: example-gw
            address: 10.0.42.1
            hcounters: yes

            name: example-gw2
            address: 10.0.43.1
            """
        )
    yield name



"""Notification channel implementation for Zino 2.0.

Notification channels are currently part of the legacy API from the Tcl-based Zino 1.0.  They are a simple text-based,
line-oriented protocol.  Clients are not expected to send any data to a notification channel, only receive data from
the server.
"""

import asyncio
import logging
from typing import TYPE_CHECKING, Any, Iterator, NamedTuple, Optional

from zino.api import auth
from zino.state import ZinoState
from zino.statemodels import Event, EventState

if TYPE_CHECKING:
    from zino.api.legacy import Zino1ServerProtocol
    from zino.api.server import ZinoServer

_logger = logging.getLogger(__name__)


class Notification(NamedTuple):
    """Represents the contents of a single notification"""

    event_id: int
    change_type: str
    value: Any


class Zino1NotificationProtocol(asyncio.Protocol):
    """Basic implementation of the Zino 1 notification protocol"""

    def __init__(self, server: Optional["ZinoServer"] = None, state: Optional[ZinoState] = None):
        """Initializes a protocol instance.

        :param server: An optional instance of `ZinoServer`.
        :param state: An optional reference to a running Zino state that this server should be based on.  If omitted,
                      this protocol will create and work on an empty state object.
        """
        self.server = server
        self.transport: Optional[asyncio.Transport] = None
        self.nonce: Optional[str] = None

        self._state = state if state is not None else ZinoState()
        self._tied_to: "Zino1ServerProtocol" = None

    @property
    def peer_name(self) -> Optional[str]:
        return self.transport.get_extra_info("peername") if self.transport else None

    def connection_made(self, transport: asyncio.Transport):
        self.transport = transport
        _logger.debug("New notification channel from %s", self.peer_name)
        self.nonce = auth.get_challenge()  # Challenges are also useful as nonces
        if self.server:
            self.server.notification_channels[self.nonce] = self
        self._respond_raw(self.nonce)

    def connection_lost(self, exc: Optional[Exception]) -> None:
        _logger.info("Lost connection from %s: %s", self.peer_name, exc)
        if self.server:
            del self.server.notification_channels[self.nonce]

    def goodbye(self):
        """Called by the tied server channel when that closes to gracefully close this channel too"""
        self._respond_raw("Normal quit from client, closing down")
        self.transport.close()

    @property
    def tied_to(self) -> Optional["Zino1ServerProtocol"]:
        return self._tied_to

    @tied_to.setter
    def tied_to(self, client: "Zino1ServerProtocol") -> None:
        self._tied_to = client

    def notify(self, notification: Notification):
        """Sends a notification to the connected client"""
        self._respond_raw(f"{notification.event_id} {notification.change_type} {notification.value or ''}")

    def _respond_raw(self, message: str):
        """Encodes and sends a response line to the connected client"""
        self.transport.write(f"{message}\r\n".encode("utf-8"))

    @classmethod
    def build_and_send_notifications(
        cls, server: "ZinoServer", new_event: Event, old_event: Optional[Event] = None
    ) -> None:
        """Prepares and sends notifications for all changes between old_event and new_event to all connected and tied
        notification channels.
        """
        notifications = list(cls.build_notifications(server.state, new_event, old_event))
        tied_channels = [channel for channel in server.notification_channels.values() if channel.tied_to]
        _logger.debug("Sending %s notifications to %s tied channels", len(notifications), len(tied_channels))

        for notification in notifications:
            for channel in tied_channels:
                channel.notify(notification)

    @classmethod
    def build_notifications(
        cls, state: ZinoState, new_event: Event, old_event: Optional[Event] = None
    ) -> Iterator[Notification]:
        """Generates a sequence of Notification objects from the changes detected between old_event and new_event.

        If `old_event` is `None`, it is assumed the event is brand new, and only the state change from EMBRYONIC
        matters.
        """
        if new_event.id not in state.events.events:
            # Event has just been scavenged after being closed for a while.  No more notifications should be sent
            # after this, since the event no longer exists.
            yield Notification(new_event.id, "scavenged", None)
            return

        changed = new_event.get_changed_fields(old_event) if old_event else ["state"]

        for attr in changed:
            if attr == "state":
                old_state = EventState.EMBRYONIC if not old_event else old_event.state
                yield Notification(new_event.id, attr, f"{old_state.value} {new_event.state.value}")

            elif attr in ("log", "history"):
                yield Notification(new_event.id, attr, 1)

            else:
                yield Notification(new_event.id, "attr", attr)



import logging
import shutil
from datetime import timedelta
from unittest.mock import Mock, patch

import pytest

from zino import flaps
from zino.config.models import PollDevice
from zino.statemodels import FlapState, InterfaceState, PortStateEvent
from zino.time import now
from zino.trapd.base import TrapMessage
from zino.trapobservers.link_traps import LinkTrapObserver

from .. import SNMPTRAP_MISSING
from ..trapd import send_trap_externally

OID_LINKDOWN = ".1.3.6.1.6.3.1.1.5.3"
OID_IFINDEX = ".1.3.6.1.2.1.2.2.1.1"
OID_IFOPERSTATUS = ".1.3.6.1.2.1.2.2.1.8"


class TestLinkTrapObserver:
    @pytest.mark.skipif(not shutil.which("snmptrap"), reason=SNMPTRAP_MISSING)
    async def test_when_link_down_is_received_it_should_create_portstate_event(
        self, state_with_localhost_with_port, localhost_pysnmp_receiver
    ):
        assert not state_with_localhost_with_port.events.get("localhost", 1, PortStateEvent), (
            "initial state should be empty"
        )

        observer = LinkTrapObserver(
            state=localhost_pysnmp_receiver.state,
            polldevs=localhost_pysnmp_receiver.polldevs,
            loop=localhost_pysnmp_receiver.loop,
        )
        localhost_pysnmp_receiver.observe(observer, *LinkTrapObserver.WANTED_TRAPS)
        await send_trap_externally(
            OID_LINKDOWN, OID_IFINDEX, "i", "1", OID_IFOPERSTATUS, "i", "2", port=localhost_pysnmp_receiver.port
        )

        assert state_with_localhost_with_port.events.get("localhost", 1, PortStateEvent), (
            "no portstate event was created"
        )

    @pytest.mark.skipif(not shutil.which("snmptrap"), reason=SNMPTRAP_MISSING)
    async def test_when_port_does_not_match_watch_pattern_it_should_ignore_link_traps(
        self, state_with_localhost_with_port, localhost_pysnmp_receiver
    ):
        assert not state_with_localhost_with_port.events.get("localhost", 1, PortStateEvent), (
            "initial state should be empty"
        )
        localhost_config = PollDevice(name="localhost", address="127.0.0.1", watchpat="foo.*")
        localhost_pysnmp_receiver.polldevs["localhost"] = localhost_config

        observer = LinkTrapObserver(
            state=localhost_pysnmp_receiver.state,
            polldevs=localhost_pysnmp_receiver.polldevs,
            loop=localhost_pysnmp_receiver.loop,
        )
        localhost_pysnmp_receiver.observe(observer, *LinkTrapObserver.WANTED_TRAPS)
        await send_trap_externally(OID_LINKDOWN, OID_IFINDEX, "i", "1", OID_IFOPERSTATUS, "i", "2")

        assert not state_with_localhost_with_port.events.get("localhost", 1, PortStateEvent), (
            "linkDown for non-watched port was not ignored"
        )

    @pytest.mark.skipif(not shutil.which("snmptrap"), reason=SNMPTRAP_MISSING)
    async def test_when_port_matches_ignore_pattern_it_should_ignore_link_traps(
        self, state_with_localhost_with_port, localhost_pysnmp_receiver
    ):
        assert not state_with_localhost_with_port.events.get("localhost", 1, PortStateEvent), (
            "initial state should be empty"
        )
        localhost_config = PollDevice(name="localhost", address="127.0.0.1", ignorepat=".*eth0.*")
        localhost_pysnmp_receiver.polldevs["localhost"] = localhost_config

        observer = LinkTrapObserver(
            state=localhost_pysnmp_receiver.state,
            polldevs=localhost_pysnmp_receiver.polldevs,
            loop=localhost_pysnmp_receiver.loop,
        )
        localhost_pysnmp_receiver.observe(observer, *LinkTrapObserver.WANTED_TRAPS)
        await send_trap_externally(OID_LINKDOWN, OID_IFINDEX, "i", "1", OID_IFOPERSTATUS, "i", "2")

        assert not state_with_localhost_with_port.events.get("localhost", 1, PortStateEvent), (
            "linkDown for non-watched port was not ignored"
        )

    def test_when_event_exists_policy_should_not_ignore_trap(self, state_with_localhost_with_port):
        observer = LinkTrapObserver(state=state_with_localhost_with_port, polldevs=Mock())
        localhost = state_with_localhost_with_port.devices.devices["localhost"]
        state_with_localhost_with_port.events.create_event(localhost.name, 1, PortStateEvent)
        assert not observer.is_port_ignored_by_policy(localhost, localhost.ports[1], is_up=False)

    def test_when_boot_time_is_unknown_policy_should_not_ignore_trap(self, state_with_localhost_with_port):
        observer = LinkTrapObserver(state=state_with_localhost_with_port, polldevs=Mock())
        localhost = state_with_localhost_with_port.devices.devices["localhost"]
        localhost.boot_time = None
        assert not observer.is_port_ignored_by_policy(localhost, localhost.ports[1], is_up=False)

    def test_when_boot_age_is_less_than_5_minutes_policy_should_ignore_trap(self, state_with_localhost_with_port):
        observer = LinkTrapObserver(state=state_with_localhost_with_port, polldevs=Mock())
        localhost = state_with_localhost_with_port.devices.devices["localhost"]
        localhost.boot_time = now() - timedelta(minutes=2)
        assert observer.is_port_ignored_by_policy(localhost, localhost.ports[1], is_up=False)

    def test_when_link_trap_is_redundant_policy_should_ignore_trap(self, state_with_localhost_with_port):
        observer = LinkTrapObserver(state=state_with_localhost_with_port, polldevs=Mock())
        localhost = state_with_localhost_with_port.devices.devices["localhost"]
        port = localhost.ports[1]
        port.state = InterfaceState.DOWN
        assert observer.is_port_ignored_by_policy(localhost, localhost.ports[1], is_up=False), (
            "did not ignore redundant linkDown trap"
        )

    async def test_when_link_trap_is_missing_ifindex_value_it_should_ignore_trap_early(
        self, state_with_localhost_with_port
    ):
        observer = LinkTrapObserver(state=state_with_localhost_with_port, polldevs=Mock())
        trap = TrapMessage(agent=Mock())
        with patch.object(observer, "handle_link_transition") as handle_link_transition:
            assert not await observer.handle_trap(trap)
            assert not handle_link_transition.called, "handle_link_transition was called"

    async def test_when_link_trap_refers_to_unknown_port_it_should_ignore_trap_early(
        self, state_with_localhost_with_port
    ):
        observer = LinkTrapObserver(state=state_with_localhost_with_port, polldevs=Mock())
        localhost = state_with_localhost_with_port.devices.devices["localhost"]
        trap = TrapMessage(agent=Mock(device=localhost), variables=[Mock(var="ifIndex", value=99)])
        with patch.object(observer, "handle_link_transition") as handle_link_transition:
            assert not await observer.handle_trap(trap)
            assert not handle_link_transition.called, "handle_link_transition was called"

    @pytest.mark.skipif(not shutil.which("snmptrap"), reason=SNMPTRAP_MISSING)
    async def test_when_event_is_new_it_should_set_lasttrans(
        self, state_with_localhost_with_port, localhost_pysnmp_receiver
    ):
        assert not state_with_localhost_with_port.events.get("localhost", 1, PortStateEvent), (
            "initial state should be empty"
        )

        observer = LinkTrapObserver(
            state=localhost_pysnmp_receiver.state,
            polldevs=localhost_pysnmp_receiver.polldevs,
            loop=localhost_pysnmp_receiver.loop,
        )
        localhost_pysnmp_receiver.observe(observer, *LinkTrapObserver.WANTED_TRAPS)
        await send_trap_externally(
            OID_LINKDOWN, OID_IFINDEX, "i", "1", OID_IFOPERSTATUS, "i", "2", port=localhost_pysnmp_receiver.port
        )

        event = state_with_localhost_with_port.events.get("localhost", 1, PortStateEvent)
        assert event.portstate == InterfaceState.DOWN
        assert event.lasttrans, "lasttrans not set"


class TestLinkTrapObserverHandleLinkTransitions:
    def test_when_port_is_ignored_by_patterns_it_should_not_create_portstate_event(
        self, monkeypatch, state_with_localhost_with_port
    ):
        observer = LinkTrapObserver(state=state_with_localhost_with_port, polldevs={})
        localhost = state_with_localhost_with_port.devices.devices["localhost"]
        port = next(iter(localhost.ports.values()))
        initial_count = len(state_with_localhost_with_port.events)
        monkeypatch.setattr(observer, "is_port_ignored_by_patterns", lambda *args, **kwargs: True)

        observer.handle_link_transition(localhost, port, is_up=False)

        assert len(state_with_localhost_with_port.events) == initial_count

    def test_when_port_is_ignored_by_policy_it_should_not_create_portstate_event(
        self, monkeypatch, state_with_localhost_with_port
    ):
        observer = LinkTrapObserver(state=state_with_localhost_with_port, polldevs={})
        localhost = state_with_localhost_with_port.devices.devices["localhost"]
        port = next(iter(localhost.ports.values()))
        initial_count = len(state_with_localhost_with_port.events)
        monkeypatch.setattr(observer, "is_port_ignored_by_policy", lambda *args, **kwargs: True)

        observer.handle_link_transition(localhost, port, is_up=False)

        assert len(state_with_localhost_with_port.events) == initial_count

    def test_when_link_is_down_it_should_create_portstate_event(self, state_with_localhost_with_port):
        observer = LinkTrapObserver(state=state_with_localhost_with_port, polldevs={})
        localhost = state_with_localhost_with_port.devices.devices["localhost"]
        port = next(iter(localhost.ports.values()))

        observer.handle_link_transition(localhost, port, is_up=False)

        event = state_with_localhost_with_port.events.get("localhost", port.ifindex, PortStateEvent)
        assert isinstance(event, PortStateEvent)
        assert event.router == "localhost"
        assert event.port == port.ifdescr

    def test_when_link_is_flapping_it_should_create_portstate_event_with_flapstate(
        self, state_with_localhost_with_port
    ):
        observer = LinkTrapObserver(state=state_with_localhost_with_port, polldevs={})
        localhost = state_with_localhost_with_port.devices.devices["localhost"]
        port = next(iter(localhost.ports.values()))
        flap = state_with_localhost_with_port.flapping.first_flap(("localhost", port.ifindex))
        flap.hist_val = flaps.FLAP_THRESHOLD * 2

        observer.handle_link_transition(localhost, port, is_up=False)

        event = state_with_localhost_with_port.events.get("localhost", port.ifindex, PortStateEvent)
        assert isinstance(event, PortStateEvent)
        assert event.router == "localhost"
        assert event.port == port.ifdescr
        assert event.flapstate == FlapState.FLAPPING

    def test_when_link_transitions_to_flapping_it_should_flag_active_flap_state(self, state_with_localhost_with_port):
        observer = LinkTrapObserver(state=state_with_localhost_with_port, polldevs={})
        localhost = state_with_localhost_with_port.devices.devices["localhost"]
        port = next(iter(localhost.ports.values()))
        flap = state_with_localhost_with_port.flapping.first_flap(("localhost", port.ifindex))
        flap.hist_val = flaps.FLAP_THRESHOLD * 2
        flap.in_active_flap_state = False

        observer.handle_link_transition(localhost, port, is_up=False)

        assert flap.in_active_flap_state

    def test_it_should_log_flap_stats_on_every_100_flaps(self, state_with_localhost_with_port, caplog):
        observer = LinkTrapObserver(state=state_with_localhost_with_port, polldevs={})
        localhost = state_with_localhost_with_port.devices.devices["localhost"]
        port = next(iter(localhost.ports.values()))
        flap = state_with_localhost_with_port.flapping.first_flap(("localhost", port.ifindex))
        flap.hist_val = flaps.FLAP_THRESHOLD * 2
        flap.in_active_flap_state = True
        flap.flaps = 99

        with caplog.at_level(logging.INFO):
            observer.handle_link_transition(localhost, port, is_up=False)

        assert "flaps, penalty" in caplog.text

    def test_when_link_transitions_to_non_flapping_it_should_set_event_flapstate_to_stable(
        self, state_with_localhost_with_port
    ):
        observer = LinkTrapObserver(state=state_with_localhost_with_port, polldevs={})
        localhost = state_with_localhost_with_port.devices.devices["localhost"]
        port = next(iter(localhost.ports.values()))
        index = ("localhost", port.ifindex)
        flap = state_with_localhost_with_port.flapping.first_flap(index)
        flap.hist_val = flaps.FLAP_MIN - 1
        flap.in_active_flap_state = True

        orig_event = state_with_localhost_with_port.events.get_or_create_event(
            "localhost", port.ifindex, PortStateEvent
        )
        orig_event.ifindex = port.ifindex
        orig_event.flapstate = FlapState.FLAPPING
        state_with_localhost_with_port.events.commit(orig_event)

        observer.handle_link_transition(localhost, port, is_up=True)

        event = state_with_localhost_with_port.events.get("localhost", port.ifindex, PortStateEvent)
        assert event.flapstate == FlapState.STABLE

    def test_when_link_transitions_to_non_flapping_it_should_remove_flap_state(self, state_with_localhost_with_port):
        observer = LinkTrapObserver(state=state_with_localhost_with_port, polldevs={})
        localhost = state_with_localhost_with_port.devices.devices["localhost"]
        port = next(iter(localhost.ports.values()))
        index = ("localhost", port.ifindex)
        flap = state_with_localhost_with_port.flapping.first_flap(index)
        flap.hist_val = flaps.FLAP_MIN - 1
        flap.in_active_flap_state = True

        orig_event = state_with_localhost_with_port.events.get_or_create_event(
            "localhost", port.ifindex, PortStateEvent
        )
        orig_event.ifindex = port.ifindex
        orig_event.flapstate = FlapState.FLAPPING
        state_with_localhost_with_port.events.commit(orig_event)

        observer.handle_link_transition(localhost, port, is_up=True)

        assert index not in state_with_localhost_with_port.flapping.interfaces

    def test_when_link_transitions_from_up_to_flapping_it_should_update_lasttrans_for_related_event(
        self, state_with_localhost_with_port
    ):
        observer = LinkTrapObserver(state=state_with_localhost_with_port, polldevs={})
        localhost = state_with_localhost_with_port.devices.devices["localhost"]
        port = next(iter(localhost.ports.values()))
        flap = state_with_localhost_with_port.flapping.first_flap(("localhost", port.ifindex))
        flap.hist_val = flaps.FLAP_THRESHOLD * 2

        initial_lasttrans = now() - timedelta(minutes=5)
        events = state_with_localhost_with_port.events
        event = events.create_event("localhost", port.ifindex, PortStateEvent)
        event.ifindex = port.ifindex
        event.portstate = InterfaceState.UP
        event.lasttrans = initial_lasttrans
        events.commit(event)

        observer.handle_link_transition(localhost, port, is_up=False)

        updated_event = events[event.id]

        assert updated_event.flapstate == FlapState.FLAPPING
        assert updated_event.portstate == InterfaceState.DOWN
        assert updated_event.lasttrans > initial_lasttrans

    async def test_when_link_transitions_from_flapping_to_up_it_should_update_lasttrans_for_related_event(
        self, state_with_localhost_with_port
    ):
        observer = LinkTrapObserver(state=state_with_localhost_with_port, polldevs={})
        localhost = state_with_localhost_with_port.devices.devices["localhost"]
        port = next(iter(localhost.ports.values()))
        index = ("localhost", port.ifindex)
        flap = state_with_localhost_with_port.flapping.first_flap(index)
        flap.hist_val = flaps.FLAP_MIN - 1
        flap.in_active_flap_state = True

        initial_lasttrans = now() - timedelta(minutes=5)
        events = state_with_localhost_with_port.events
        event = events.create_event("localhost", port.ifindex, PortStateEvent)
        event.ifindex = port.ifindex
        event.flapstate = FlapState.FLAPPING
        event.portstate = InterfaceState.DOWN
        event.lasttrans = initial_lasttrans
        events.commit(event)

        observer.handle_link_transition(localhost, port, is_up=True)

        updated_event = events[event.id]
        assert updated_event.flapstate == FlapState.STABLE
        assert updated_event.portstate == InterfaceState.UP
        assert updated_event.lasttrans > initial_lasttrans

    async def test_when_link_transitions_from_flapping_to_up_it_should_update_ac_down_for_related_event(
        self, state_with_localhost_with_port
    ):
        observer = LinkTrapObserver(state=state_with_localhost_with_port, polldevs={})
        localhost = state_with_localhost_with_port.devices.devices["localhost"]
        port = next(iter(localhost.ports.values()))
        index = ("localhost", port.ifindex)
        flap = state_with_localhost_with_port.flapping.first_flap(index)
        flap.hist_val = flaps.FLAP_MIN - 1
        flap.in_active_flap_state = True

        initial_lasttrans = now() - timedelta(minutes=5)
        initial_ac_down = timedelta(0)
        events = state_with_localhost_with_port.events
        event = events.create_event("localhost", port.ifindex, PortStateEvent)
        event.ifindex = port.ifindex
        event.flapstate = FlapState.FLAPPING
        event.portstate = InterfaceState.DOWN
        event.ac_down = initial_ac_down
        event.lasttrans = initial_lasttrans
        events.commit(event)

        observer.handle_link_transition(localhost, port, is_up=True)

        updated_event = events[event.id]
        assert updated_event.flapstate == FlapState.STABLE
        assert updated_event.portstate == InterfaceState.UP
        assert updated_event.ac_down > initial_ac_down


class TestLinkTrapObserverIsPortIgnoredByPatterns:
    async def test_when_port_matches_watch_pattern_it_should_return_false(
        self, state_with_localhost_with_port, localhost_pysnmp_receiver
    ):
        localhost_config = PollDevice(name="localhost", address="127.0.0.1", watchpat="bar")
        localhost_pysnmp_receiver.polldevs["localhost"] = localhost_config

        observer = LinkTrapObserver(
            state=localhost_pysnmp_receiver.state,
            polldevs=localhost_pysnmp_receiver.polldevs,
            loop=localhost_pysnmp_receiver.loop,
        )
        device = state_with_localhost_with_port.devices["localhost"]
        assert not observer.is_port_ignored_by_patterns(device, "foobar")

    async def test_when_port_matches_ignore_pattern_it_should_return_true(
        self, state_with_localhost_with_port, localhost_pysnmp_receiver
    ):
        localhost_config = PollDevice(name="localhost", address="127.0.0.1", ignorepat="bar")
        localhost_pysnmp_receiver.polldevs["localhost"] = localhost_config

        observer = LinkTrapObserver(
            state=localhost_pysnmp_receiver.state,
            polldevs=localhost_pysnmp_receiver.polldevs,
            loop=localhost_pysnmp_receiver.loop,
        )
        device = state_with_localhost_with_port.devices["localhost"]
        assert observer.is_port_ignored_by_patterns(device, "foobar")



import logging

from apscheduler.jobstores.base import JobLookupError

from zino.scheduler import get_scheduler
from zino.statemodels import ReachabilityEvent, ReachabilityState
from zino.tasks.errors import DeviceUnreachableError
from zino.tasks.task import Task

_logger = logging.getLogger(__name__)


class ReachableTask(Task):
    EXTRA_JOB_INTERVAL = 60

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._scheduler = get_scheduler()

    async def run(self):
        """Checks if device is reachable. Schedules extra reachability checks if not."""
        if self._extra_job_is_running():
            raise DeviceUnreachableError
        try:
            await self._get_uptime()
        except TimeoutError:
            event = self.state.events.get_or_create_event(self.device.name, None, ReachabilityEvent)
            if event.reachability != ReachabilityState.NORESPONSE:
                event.reachability = ReachabilityState.NORESPONSE
                if self.device_state.reachable_in_last_run is None and self.config.event.make_events_for_new_devices:
                    event.add_log(f"New device: {self.device.name} no-response")
                else:
                    event.add_log(f"{self.device.name} no-response")
            event.polladdr = self.device.address
            event.priority = self.device.priority
            self.state.events.commit(event)
            self.device_state.reachable_in_last_run = False
            self._schedule_extra_job()
            raise DeviceUnreachableError
        else:
            _logger.debug("Device %s is reachable", self.device.name)
            self._update_reachability_event_as_reachable()
            if self.device_state.reachable_in_last_run is None and self.config.event.make_events_for_new_devices:
                self._post_reachability_reachable_event_for_new_device()
            self.device_state.reachable_in_last_run = True

    async def _run_extra_job(self):
        try:
            # This runs outside a job context, so we need to ensure we clean up low-level SNMP resources
            with self.snmp:
                await self._get_uptime()
        except TimeoutError:
            return
        else:
            _logger.debug("Device %s is reachable", self.device.name)
            self._update_reachability_event_as_reachable()
            self._deschedule_extra_job()

    def _update_reachability_event_as_reachable(self):
        event = self.state.events.get(self.device.name, None, ReachabilityEvent)
        if event and event.reachability != ReachabilityState.REACHABLE:
            event = self.state.events.checkout(event.id)
            event.add_log(f"{self.device.name} reachable")
            event.reachability = ReachabilityState.REACHABLE
            self.state.events.commit(event)

    def _schedule_extra_job(self):
        name = self._get_extra_job_name()
        self._scheduler.add_job(
            func=self._run_extra_job,
            trigger="interval",
            seconds=self.EXTRA_JOB_INTERVAL,
            name=name,
            id=name,
        )

    def _deschedule_extra_job(self):
        name = self._get_extra_job_name()
        try:
            self._scheduler.remove_job(job_id=name)
        except JobLookupError:
            pass

    def _extra_job_is_running(self):
        name = self._get_extra_job_name()
        if self._scheduler.get_job(job_id=name):
            return True
        else:
            return False

    def _get_extra_job_name(self):
        return f"reachabletask_{self.device.name}"

    def _post_reachability_reachable_event_for_new_device(self):
        event = self.state.events.create_event(self.device.name, None, ReachabilityEvent)
        event.add_log(f"New device: {self.device.name} reachable")
        event.reachability = ReachabilityState.REACHABLE
        self.state.events.commit(event)



import logging
from datetime import datetime, timezone

from zino.flaps import FlappingState
from zino.state import ZinoState
from zino.stateconverter.linedata import LineData
from zino.stateconverter.utils import OldState

_log = logging.getLogger(__name__)


def set_flapping_state(old_state: OldState, new_state: ZinoState):
    for linedata in old_state["::firstFlap"]:
        _set_first_flap(linedata, new_state)
    for linedata in old_state["::flapHistVal"]:
        _set_flap_hist_val(linedata, new_state)
    for linedata in old_state["::flappedAboveThreshold"]:
        _set_flapped_above_threshold(linedata, new_state)
    for linedata in old_state["::flapping"]:
        _set_flapping(linedata, new_state)
    for linedata in old_state["::flaps"]:
        _set_flaps(linedata, new_state)
    for linedata in old_state["::lastFlap"]:
        _set_last_flap(linedata, new_state)
    for linedata in old_state["::lastAge"]:
        _set_last_age(linedata, new_state)


def _set_first_flap(linedata: LineData, state: ZinoState):
    timestamp = datetime.fromtimestamp(int(linedata.value), tz=timezone.utc)
    _get_flapping_state_for(linedata, state).first_flap = timestamp


def _set_flap_hist_val(linedata: LineData, state: ZinoState):
    _get_flapping_state_for(linedata, state).hist_val = float(linedata.value)


def _set_flapped_above_threshold(linedata: LineData, state: ZinoState):
    _get_flapping_state_for(linedata, state).flapped_above_threshold = bool(int(linedata.value))


def _set_flapping(linedata: LineData, state: ZinoState):
    # This really isn't a persisted attribute in Zino 2, it is always calculated on the fly
    pass


def _set_flaps(linedata: LineData, state: ZinoState):
    _get_flapping_state_for(linedata, state).flaps = int(linedata.value)


def _set_last_flap(linedata: LineData, state: ZinoState):
    timestamp = datetime.fromtimestamp(int(linedata.value), tz=timezone.utc)
    _get_flapping_state_for(linedata, state).last_flap = timestamp


def _set_last_age(linedata: LineData, state: ZinoState):
    timestamp = datetime.fromtimestamp(int(linedata.value), tz=timezone.utc)
    _get_flapping_state_for(linedata, state).last_age = timestamp


def _get_flapping_state_for(linedata: LineData, state: ZinoState) -> FlappingState:
    device, port = linedata.identifiers
    index = (device, int(port))
    return state.flapping.interfaces.setdefault(index, FlappingState())



from unittest.mock import patch

import pytest

from zino.config.models import PollDevice
from zino.state import ZinoState
from zino.tasks.vendor import VendorTask


class TestVendorTask:
    async def test_run_should_set_enterprise_id(self, snmpsim, snmp_test_port):
        device = PollDevice(name="localhost", address="127.0.0.1", community="public", port=snmp_test_port)
        state = ZinoState()
        task = VendorTask(device, state)
        assert (await task.run()) is None
        assert device.name in state.devices
        # The "public" test fixture is from an HP switch
        assert state.devices[device.name].enterprise_id == 11

    async def test_run_should_raise_exception_when_there_is_no_response(self):
        device = PollDevice(name="localhost", address="127.0.0.1", community="invalid", port=666)
        state = ZinoState()
        task = VendorTask(device, state)
        with patch("zino.snmp.SNMP.get") as get_mock:
            get_mock.side_effect = TimeoutError
            with pytest.raises(TimeoutError):
                await task.run()
        assert len(state.devices) == 0



from ipaddress import IPv4Address

import pytest

from zino.config.models import PollDevice
from zino.oid import OID
from zino.state import ZinoState
from zino.statemodels import BFDEvent, BFDSessState, BFDState, Port
from zino.tasks.bfdtask import BFDTask


class TestJuniper:
    @pytest.mark.parametrize("task", ["juniper-bfd-up"], indirect=True)
    def test_parse_row_creates_correct_bfd_state(self, task, bfd_state):
        state = task._parse_row(
            OID(f".{bfd_state.session_index}"),
            "up",
            bfd_state.session_discr,
            IPv4Address("127.0.0.1"),
        )
        assert state == bfd_state

    @pytest.mark.parametrize("task", ["juniper-bfd-up"], indirect=True)
    async def test_poll_juniper_returns_correct_ifdescr_to_state_mapping(self, task, bfd_state, device_port):
        result = await task._poll_juniper()
        assert device_port.ifdescr in result
        state = result.get(device_port.ifdescr)
        assert state == bfd_state

    @pytest.mark.parametrize("task", ["juniper-bfd-up"], indirect=True)
    async def test_when_single_index_is_given_then_poll_juniper_should_return_correct_ifdescr_to_state_mapping(
        self, task, bfd_state, device_port
    ):
        result = await task._poll_juniper(session_index=bfd_state.session_index)
        assert device_port.ifdescr in result
        state = result.get(device_port.ifdescr)
        assert state == bfd_state


@pytest.mark.parametrize("task", ["juniper-bfd-up", "cisco-bfd-up"], indirect=True)
async def test_task_updates_state_correctly(task, device_port, bfd_state):
    assert not device_port.bfd_state
    await task.run()
    assert device_port.bfd_state == bfd_state


@pytest.mark.parametrize("task", ["juniper-bfd-down", "cisco-bfd-down"], indirect=True)
async def test_event_should_be_made_the_first_time_a_port_is_polled_if_state_is_not_up(task, device_port):
    assert not device_port.bfd_state
    await task.run()
    assert device_port.bfd_state
    event = task.state.events.get(
        device_name=task.device.name,
        subindex=device_port.ifindex,
        event_class=BFDEvent,
    )
    assert event


@pytest.mark.parametrize("task", ["juniper-bfd-up", "cisco-bfd-up"], indirect=True)
async def test_event_should_not_be_made_the_first_time_a_port_is_polled_if_state_is_up(task, device_port):
    assert not device_port.bfd_state
    await task.run()
    assert device_port.bfd_state
    event = task.state.events.get(
        device_name=task.device.name,
        subindex=device_port.ifindex,
        event_class=BFDEvent,
    )
    assert not event


@pytest.mark.parametrize("task", ["juniper-bfd-up", "cisco-bfd-up"], indirect=True)
async def test_state_changing_should_create_event(task, device_port, bfd_state):
    down_state = BFDState(session_state=BFDSessState.DOWN, session_index=bfd_state.session_index)
    device_port.bfd_state = down_state
    await task.run()
    assert device_port.bfd_state != down_state
    event = task.state.events.get(device_name=task.device.name, subindex=device_port.ifindex, event_class=BFDEvent)
    assert event


@pytest.mark.parametrize("task", ["juniper-bfd-up", "cisco-bfd-up"], indirect=True)
async def test_state_not_changing_should_not_create_event(task, device_port, bfd_state):
    device_port.bfd_state = bfd_state
    await task.run()
    assert device_port.bfd_state == bfd_state
    event = task.state.events.get(device_name=task.device.name, subindex=device_port.ifindex, event_class=BFDEvent)
    assert not event


@pytest.fixture()
def device_port():
    """Port related to the BFD state at the simulated device"""
    yield Port(ifindex=1, ifdescr="xe-1/2/0.0")


@pytest.fixture()
def bfd_state():
    """Represents the BFD state at the simulated device"""
    yield BFDState(
        session_state=BFDSessState.UP,
        session_index=4524,
        session_discr=4524,
        session_addr=IPv4Address("127.0.0.1"),
    )


@pytest.fixture(scope="function")
def task(request, snmpsim, snmp_test_port, device_port):
    device = PollDevice(
        name="buick.lab.example.org",
        address="127.0.0.1",
        port=snmp_test_port,
        community=request.param,
    )
    state = ZinoState()
    device_state = state.devices.get(device_name=device.name)
    if "cisco" in request.param:
        device_state.enterprise_id = 9
    elif "juniper" in request.param:
        device_state.enterprise_id = 2636
    # this effectively makes the device_port fixture a shortcut for accessing
    # the port that will change bfd_state when the task is run
    device_state.ports[device_port.ifindex] = device_port
    task = BFDTask(device, state)
    yield task



import datetime
import json
import os

import pytest

from zino.statemodels import (
    DeviceState,
    DeviceStates,
    Event,
    EventState,
    LogEntry,
    ReachabilityEvent,
    regex_search,
)
from zino.time import now


class TestEvent:
    def test_add_log_should_set_proper_timestamp(self, fake_event):
        log = fake_event.add_log("test")
        assert isinstance(log.timestamp, datetime.datetime)

    def test_add_history_should_set_proper_timestamp(self, fake_event):
        log = fake_event.add_history("test")
        assert isinstance(log.timestamp, datetime.datetime)

    def test_model_dump_simple_attrs_should_return_dict_of_only_string_values(self, fake_event):
        attrs = fake_event.model_dump_simple_attrs()

        assert all(isinstance(val, str) for val in attrs.values())

    def test_model_dump_simple_attrs_should_not_return_dict_with_log_or_history(self, fake_event):
        attrs = fake_event.model_dump_simple_attrs()

        assert "log" not in attrs
        assert "history" not in attrs

    def test_model_dump_simple_attrs_should_replace_underscores_with_dashes_in_attr_names(self, fake_event):
        """Legacy Zino protocol event attributes use dashes, not underscores in their names."""
        fake_event.ac_down = datetime.timedelta(seconds=42)
        attrs = fake_event.model_dump_simple_attrs()

        assert "ac_down" not in attrs
        assert "ac-down" in attrs

    def test_model_dump_simple_attrs_should_represent_timedeltas_as_number_of_seconds(self, fake_event):
        """Legacy Zino protocol event attributes use dashes, not underscores in their names."""
        fake_event.ac_down = datetime.timedelta(seconds=42)
        attrs = fake_event.model_dump_simple_attrs()

        assert attrs["ac-down"] == "42"

    def test_zinoify_value_when_value_is_enum_it_should_return_its_real_value(self):
        assert Event.zinoify_value(EventState.OPEN) == "open"

    def test_zinoify_value_when_value_is_int_it_should_return_a_str(self):
        assert Event.zinoify_value(42) == "42"

    def test_zinoify_value_when_value_is_datetime_it_should_return_a_unix_timestamp_as_a_str(self):
        timestamp = datetime.datetime(2023, 10, 18, 11, 53, 37, tzinfo=datetime.timezone.utc)
        assert Event.zinoify_value(timestamp) == "1697630017"

    def test_zinoify_value_when_value_is_anything_its_str_should_be_returned(self):
        class Throwaway:
            def __str__(self):
                return "foo"

        assert Event.zinoify_value(Throwaway()) == "foo"

    def test_when_set_state_is_called_it_should_change_state(self, fake_event):
        fake_event.set_state(EventState.CLOSED, user="nobody")
        assert fake_event.state == EventState.CLOSED

    def test_when_state_is_changed_set_state_should_add_history_entry(self, fake_event):
        history_count_before = len(fake_event.history)
        fake_event.set_state(EventState.CLOSED, user="nobody")
        history_count_after = len(fake_event.history)
        assert history_count_after > history_count_before

    def test_when_state_is_changed_set_state_should_add_history_entry_with_details(self, fake_event):
        old_state = fake_event.state
        fake_event.set_state(EventState.CLOSED, user="zaphod")
        last_history_entry = fake_event.history[-1]

        for detail in (old_state.value, EventState.CLOSED.value, "zaphod"):
            assert detail in last_history_entry.message

    def test_when_state_is_unchanged_set_state_should_not_add_history_entry(self, fake_event):
        history_count_before = len(fake_event.history)
        fake_event.set_state(fake_event.state, user="nobody")
        history_count_after = len(fake_event.history)
        assert history_count_after == history_count_before

    def test_get_changed_fields_should_correctly_detect_changed_fields(self, fake_event):
        copy = fake_event.model_copy(deep=True)
        copy.add_log("test")
        copy.updated = now() + datetime.timedelta(seconds=3)

        changed = fake_event.get_changed_fields(copy)
        assert set(changed) == {"log", "updated"}

    def test_dump_event_to_file_should_dump_valid_json_to_file(self, tmp_path, fake_event):
        fake_event.set_state(EventState.CLOSED)
        fake_event.dump_event_to_file(tmp_path)

        dumpfile = f"{tmp_path}/{fake_event.id}.json"
        assert os.path.exists(dumpfile)
        with open(dumpfile, "r") as data:
            assert json.load(data)


class TestLogEntryModelDumpLegacy:
    def test_should_be_prefixed_by_timestamp(self):
        entry = LogEntry(message="foobar")
        lines = entry.model_dump_legacy()
        timestamp, message = lines[0].split(" ")

        assert timestamp.isdigit()
        assert message == "foobar"

    def test_when_message_is_single_line_it_should_return_a_single_line(self):
        entry = LogEntry(message="foobar")
        lines = entry.model_dump_legacy()

        assert len(lines) == 1

    def test_when_message_is_multi_line_it_should_return_the_correct_number_of_lines(self):
        entry = LogEntry(message="one\ntwo\nthree")
        lines = entry.model_dump_legacy()

        assert len(lines) == 3

    def test_when_message_is_multi_line_continuation_lines_should_be_prefixed_by_space(self):
        entry = LogEntry(message="one\ntwo\nthree")
        lines = entry.model_dump_legacy()

        assert all(line.startswith(" ") for line in lines[1:])


class TestDeviceState:
    @pytest.mark.parametrize(
        "enterprise_id,property_name,expected",
        [(9, "is_cisco", True), (666, "is_cisco", False), (2636, "is_juniper", True), (666, "is_juniper", False)],
    )
    def test_vendor_utility_property_returns_expected_result(self, enterprise_id, property_name, expected):
        dev = DeviceState(name="foo", enterprise_id=enterprise_id)
        assert getattr(dev, property_name) == expected


class TestDeviceStates:
    def test_empty_dict_should_not_contain_devices(self):
        states = DeviceStates()
        assert len(states) == 0
        assert "foo" not in states

    def test_get_should_create_new_device_state(self):
        states = DeviceStates()
        router = "new"
        assert router not in states

        result = states.get(router)
        assert isinstance(result, DeviceState)
        assert router in states


class TestRegexSearch:
    def test_when_string_matches_not_at_the_beginning_it_should_return_true(self):
        assert regex_search("device", "blabla_device")

    def test_when_string_matches_at_the_beginning_it_should_return_true(self):
        assert regex_search("device", "device")

    def test_when_string_does_not_match_it_should_return_false(self):
        assert not regex_search("device", "wrong")


@pytest.fixture
def fake_event() -> ReachabilityEvent:
    return ReachabilityEvent(id=42, router="example-gw.example.org", state=EventState.OPEN)



#
# PySNMP MIB module CISCO-PIM-MIB (http://snmplabs.com/pysmi)
# ASN.1 source file:///home/mvold/d/zino/mibs/CISCO-PIM-MIB.my
# Produced by pysmi-0.3.4 at Thu Jul 18 12:37:03 2024
# On host agrajag platform Linux version 6.6.37 by user mvold
# Using Python version 3.11.9 (main, Apr  2 2024, 08:25:04) [GCC 13.2.0]
#
ObjectIdentifier, OctetString, Integer = mibBuilder.importSymbols("ASN1", "ObjectIdentifier", "OctetString", "Integer")
NamedValues, = mibBuilder.importSymbols("ASN1-ENUMERATION", "NamedValues")
ConstraintsIntersection, ConstraintsUnion, SingleValueConstraint, ValueSizeConstraint, ValueRangeConstraint = mibBuilder.importSymbols("ASN1-REFINEMENT", "ConstraintsIntersection", "ConstraintsUnion", "SingleValueConstraint", "ValueSizeConstraint", "ValueRangeConstraint")
ciscoMgmt, = mibBuilder.importSymbols("CISCO-SMI", "ciscoMgmt")
InetAddress, InetAddressType = mibBuilder.importSymbols("INET-ADDRESS-MIB", "InetAddress", "InetAddressType")
pimInterfaceStatus, pimRPSetHoldTime = mibBuilder.importSymbols("PIM-MIB", "pimInterfaceStatus", "pimRPSetHoldTime")
ModuleCompliance, ObjectGroup, NotificationGroup = mibBuilder.importSymbols("SNMPv2-CONF", "ModuleCompliance", "ObjectGroup", "NotificationGroup")
NotificationType, Counter32, Integer32, MibScalar, MibTable, MibTableRow, MibTableColumn, iso, Gauge32, IpAddress, Unsigned32, ObjectIdentity, TimeTicks, Counter64, MibIdentifier, ModuleIdentity, Bits = mibBuilder.importSymbols("SNMPv2-SMI", "NotificationType", "Counter32", "Integer32", "MibScalar", "MibTable", "MibTableRow", "MibTableColumn", "iso", "Gauge32", "IpAddress", "Unsigned32", "ObjectIdentity", "TimeTicks", "Counter64", "MibIdentifier", "ModuleIdentity", "Bits")
DisplayString, TextualConvention = mibBuilder.importSymbols("SNMPv2-TC", "DisplayString", "TextualConvention")
ciscoPimMIB = ModuleIdentity((1, 3, 6, 1, 4, 1, 9, 9, 184))
ciscoPimMIB.setRevisions(('2000-11-02 00:00',))
if mibBuilder.loadTexts: ciscoPimMIB.setLastUpdated('200011020000Z')
if mibBuilder.loadTexts: ciscoPimMIB.setOrganization('Cisco Systems, Inc.')
ciscoPimMIBObjects = MibIdentifier((1, 3, 6, 1, 4, 1, 9, 9, 184, 1))
cpim = MibIdentifier((1, 3, 6, 1, 4, 1, 9, 9, 184, 1, 1))
cpimInvalidRegisterMsgsRcvd = MibScalar((1, 3, 6, 1, 4, 1, 9, 9, 184, 1, 1, 1), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: cpimInvalidRegisterMsgsRcvd.setStatus('current')
cpimInvalidJoinPruneMsgsRcvd = MibScalar((1, 3, 6, 1, 4, 1, 9, 9, 184, 1, 1, 2), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: cpimInvalidJoinPruneMsgsRcvd.setStatus('current')
cpimLastErrorType = MibScalar((1, 3, 6, 1, 4, 1, 9, 9, 184, 1, 1, 3), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(1, 2, 3))).clone(namedValues=NamedValues(("none", 1), ("invalidRegister", 2), ("invalidJoinPrune", 3)))).setMaxAccess("readonly")
if mibBuilder.loadTexts: cpimLastErrorType.setStatus('current')
cpimLastErrorOriginType = MibScalar((1, 3, 6, 1, 4, 1, 9, 9, 184, 1, 1, 4), InetAddressType()).setMaxAccess("readonly")
if mibBuilder.loadTexts: cpimLastErrorOriginType.setStatus('current')
cpimLastErrorOrigin = MibScalar((1, 3, 6, 1, 4, 1, 9, 9, 184, 1, 1, 5), InetAddress()).setMaxAccess("readonly")
if mibBuilder.loadTexts: cpimLastErrorOrigin.setStatus('current')
cpimLastErrorGroupType = MibScalar((1, 3, 6, 1, 4, 1, 9, 9, 184, 1, 1, 6), InetAddressType()).setMaxAccess("readonly")
if mibBuilder.loadTexts: cpimLastErrorGroupType.setStatus('current')
cpimLastErrorGroup = MibScalar((1, 3, 6, 1, 4, 1, 9, 9, 184, 1, 1, 7), InetAddress()).setMaxAccess("readonly")
if mibBuilder.loadTexts: cpimLastErrorGroup.setStatus('current')
cpimLastErrorRPType = MibScalar((1, 3, 6, 1, 4, 1, 9, 9, 184, 1, 1, 8), InetAddressType()).setMaxAccess("readonly")
if mibBuilder.loadTexts: cpimLastErrorRPType.setStatus('current')
cpimLastErrorRP = MibScalar((1, 3, 6, 1, 4, 1, 9, 9, 184, 1, 1, 9), InetAddress()).setMaxAccess("readonly")
if mibBuilder.loadTexts: cpimLastErrorRP.setStatus('current')
ciscoPimMIBNotificationPrefix = MibIdentifier((1, 3, 6, 1, 4, 1, 9, 9, 184, 2))
ciscoPimMIBNotifications = MibIdentifier((1, 3, 6, 1, 4, 1, 9, 9, 184, 2, 0))
ciscoPimMIBNotificationObjects = MibIdentifier((1, 3, 6, 1, 4, 1, 9, 9, 184, 2, 1))
cpimRPMappingChangeType = MibScalar((1, 3, 6, 1, 4, 1, 9, 9, 184, 2, 1, 1), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(1, 2, 3, 4))).clone(namedValues=NamedValues(("newMapping", 1), ("deletedMapping", 2), ("modifiedOldMapping", 3), ("modifiedNewMapping", 4)))).setMaxAccess("readonly")
if mibBuilder.loadTexts: cpimRPMappingChangeType.setStatus('current')
ciscoPimInterfaceUp = NotificationType((1, 3, 6, 1, 4, 1, 9, 9, 184, 2, 0, 1)).setObjects(("PIM-MIB", "pimInterfaceStatus"))
if mibBuilder.loadTexts: ciscoPimInterfaceUp.setStatus('current')
ciscoPimInterfaceDown = NotificationType((1, 3, 6, 1, 4, 1, 9, 9, 184, 2, 0, 2)).setObjects(("PIM-MIB", "pimInterfaceStatus"))
if mibBuilder.loadTexts: ciscoPimInterfaceDown.setStatus('current')
ciscoPimRPMappingChange = NotificationType((1, 3, 6, 1, 4, 1, 9, 9, 184, 2, 0, 3)).setObjects(("PIM-MIB", "pimRPSetHoldTime"), ("CISCO-PIM-MIB", "cpimRPMappingChangeType"))
if mibBuilder.loadTexts: ciscoPimRPMappingChange.setStatus('current')
ciscoPimInvalidRegister = NotificationType((1, 3, 6, 1, 4, 1, 9, 9, 184, 2, 0, 4)).setObjects(("CISCO-PIM-MIB", "cpimLastErrorOriginType"), ("CISCO-PIM-MIB", "cpimLastErrorOrigin"), ("CISCO-PIM-MIB", "cpimLastErrorGroupType"), ("CISCO-PIM-MIB", "cpimLastErrorGroup"), ("CISCO-PIM-MIB", "cpimLastErrorRPType"), ("CISCO-PIM-MIB", "cpimLastErrorRP"), ("CISCO-PIM-MIB", "cpimInvalidRegisterMsgsRcvd"))
if mibBuilder.loadTexts: ciscoPimInvalidRegister.setStatus('current')
ciscoPimInvalidJoinPrune = NotificationType((1, 3, 6, 1, 4, 1, 9, 9, 184, 2, 0, 5)).setObjects(("CISCO-PIM-MIB", "cpimLastErrorOriginType"), ("CISCO-PIM-MIB", "cpimLastErrorOrigin"), ("CISCO-PIM-MIB", "cpimLastErrorGroupType"), ("CISCO-PIM-MIB", "cpimLastErrorGroup"), ("CISCO-PIM-MIB", "cpimLastErrorRPType"), ("CISCO-PIM-MIB", "cpimLastErrorRP"), ("CISCO-PIM-MIB", "cpimInvalidJoinPruneMsgsRcvd"))
if mibBuilder.loadTexts: ciscoPimInvalidJoinPrune.setStatus('current')
ciscoPimMIBConformance = MibIdentifier((1, 3, 6, 1, 4, 1, 9, 9, 184, 3))
ciscoPimMIBCompliances = MibIdentifier((1, 3, 6, 1, 4, 1, 9, 9, 184, 3, 1))
ciscoPimMIBGroups = MibIdentifier((1, 3, 6, 1, 4, 1, 9, 9, 184, 3, 2))
ciscoPimSparseMIBCompliance = ModuleCompliance((1, 3, 6, 1, 4, 1, 9, 9, 184, 3, 1, 1)).setObjects(("CISCO-PIM-MIB", "ciscoPimSparseMIBGroup"), ("CISCO-PIM-MIB", "ciscoPimNotificationObjectGroup"))

if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    ciscoPimSparseMIBCompliance = ciscoPimSparseMIBCompliance.setStatus('current')
ciscoPimSparseMIBGroup = ObjectGroup((1, 3, 6, 1, 4, 1, 9, 9, 184, 3, 2, 1)).setObjects(("CISCO-PIM-MIB", "cpimInvalidRegisterMsgsRcvd"), ("CISCO-PIM-MIB", "cpimInvalidJoinPruneMsgsRcvd"), ("CISCO-PIM-MIB", "cpimLastErrorType"), ("CISCO-PIM-MIB", "cpimLastErrorOriginType"), ("CISCO-PIM-MIB", "cpimLastErrorOrigin"), ("CISCO-PIM-MIB", "cpimLastErrorGroupType"), ("CISCO-PIM-MIB", "cpimLastErrorGroup"), ("CISCO-PIM-MIB", "cpimLastErrorRPType"), ("CISCO-PIM-MIB", "cpimLastErrorRP"))
if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    ciscoPimSparseMIBGroup = ciscoPimSparseMIBGroup.setStatus('current')
ciscoPimNotificationObjectGroup = ObjectGroup((1, 3, 6, 1, 4, 1, 9, 9, 184, 3, 2, 2)).setObjects(("CISCO-PIM-MIB", "cpimRPMappingChangeType"))
if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    ciscoPimNotificationObjectGroup = ciscoPimNotificationObjectGroup.setStatus('current')
mibBuilder.exportSymbols("CISCO-PIM-MIB", cpimInvalidRegisterMsgsRcvd=cpimInvalidRegisterMsgsRcvd, ciscoPimInterfaceUp=ciscoPimInterfaceUp, cpimLastErrorGroupType=cpimLastErrorGroupType, cpimInvalidJoinPruneMsgsRcvd=cpimInvalidJoinPruneMsgsRcvd, ciscoPimRPMappingChange=ciscoPimRPMappingChange, ciscoPimMIB=ciscoPimMIB, ciscoPimSparseMIBCompliance=ciscoPimSparseMIBCompliance, cpimLastErrorRP=cpimLastErrorRP, cpimLastErrorType=cpimLastErrorType, cpimRPMappingChangeType=cpimRPMappingChangeType, ciscoPimMIBNotificationPrefix=ciscoPimMIBNotificationPrefix, ciscoPimMIBConformance=ciscoPimMIBConformance, cpimLastErrorRPType=cpimLastErrorRPType, ciscoPimMIBCompliances=ciscoPimMIBCompliances, ciscoPimMIBObjects=ciscoPimMIBObjects, cpimLastErrorGroup=cpimLastErrorGroup, cpim=cpim, ciscoPimInvalidJoinPrune=ciscoPimInvalidJoinPrune, ciscoPimMIBNotifications=ciscoPimMIBNotifications, ciscoPimInvalidRegister=ciscoPimInvalidRegister, PYSNMP_MODULE_ID=ciscoPimMIB, ciscoPimMIBGroups=ciscoPimMIBGroups, ciscoPimInterfaceDown=ciscoPimInterfaceDown, cpimLastErrorOriginType=cpimLastErrorOriginType, ciscoPimSparseMIBGroup=ciscoPimSparseMIBGroup, cpimLastErrorOrigin=cpimLastErrorOrigin, ciscoPimNotificationObjectGroup=ciscoPimNotificationObjectGroup, ciscoPimMIBNotificationObjects=ciscoPimMIBNotificationObjects)



#!/usr/bin/env python3
import argparse
import asyncio
import errno
import gc
import grp
import logging
import logging.config
import os
import pwd
import sys
from asyncio import AbstractEventLoop
from collections import defaultdict
from datetime import datetime, timedelta
from typing import Any, Optional

import tzlocal
from pydantic import ValidationError

from zino import flaps, state
from zino.api.server import ZinoServer
from zino.config import InvalidConfigurationError, read_configuration
from zino.job_tracker import get_job_tracker
from zino.scheduler import get_scheduler, load_and_schedule_polldevs
from zino.snmp import import_snmp_backend
from zino.snmp.agent import ZinoSnmpAgent
from zino.statemodels import Event
from zino.trapd import import_trap_backend

# ensure all our desired trap observers are loaded.  They will not be explicitly referenced here, hence the noqa tag
from zino.trapobservers import (  # noqa
    bfd_traps,
    bgp_traps,
    ignored_traps,
    link_traps,
    logged_traps,
)
from zino.utils import file_is_world_readable

STATE_DUMP_JOB_ID = "zino.dump_state"
# Never try to dump state more often than this:
MINIMUM_STATE_DUMP_INTERVAL = timedelta(seconds=10)
DEFAULT_CONFIG_FILE = "zino.toml"
_log = logging.getLogger("zino")


def main():
    args = parse_args()
    logging.basicConfig(
        level=logging.INFO if not args.debug else logging.DEBUG,
        format="%(asctime)s - %(levelname)s - %(name)s (%(threadName)s) - %(message)s",
    )
    _config = load_config(args)
    if _config:
        state.config = _config
    apply_logging_config(state.config.logging)

    try:
        secrets_file = state.config.authentication.file
        if file_is_world_readable(secrets_file):
            _log.warning(
                f"Secrets file {secrets_file} is world-readable. Please ensure that it is only readable by the user that runs the zino process."
            )
    except OSError as e:
        _log.fatal(e)
        sys.exit(1)

    # Load the same SNMP and trap back-ends
    snmp_backend = import_snmp_backend(state.config.snmp.backend)
    snmp_backend.init_backend()
    import_trap_backend(state.config.snmp.backend)

    state.state = state.ZinoState.load_state_from_file(state.config.persistence.file) or state.ZinoState()
    state.clean_state(state.state)
    init_event_loop(args)


def load_config(args: argparse.Namespace) -> Optional[state.Configuration]:
    """
    Loads the configuration file, exiting the process if there are config errors

    Returns the configuration specified by the config file and None if no config file
    name was specified as argument and no default config file exists
    """
    try:
        return read_configuration(args.config_file or DEFAULT_CONFIG_FILE, args.polldevs)
    except OSError:
        if args.config_file:
            _log.fatal(f"No config file with the name {args.config_file} found.")
            sys.exit(1)
    except InvalidConfigurationError:
        _log.fatal(f"Configuration file with the name {args.config_file or DEFAULT_CONFIG_FILE} is invalid TOML.")
        sys.exit(1)
    except ValidationError as e:
        _log.fatal(e)
        sys.exit(1)


def apply_logging_config(logging_config: dict[str, Any]) -> None:
    """Applies the logging configuration, exiting the process if there are config errors"""
    try:
        logging.config.dictConfig(logging_config)
    except ValueError as error:
        _log.fatal(f"Invalid logging configuration: {error}")
        sys.exit(1)


def init_event_loop(args: argparse.Namespace, loop: Optional[AbstractEventLoop] = None):
    if not loop:
        loop = asyncio.get_event_loop()

    if args.trap_port:
        trap_backend = import_trap_backend()
        trap_receiver = trap_backend.TrapReceiver(port=args.trap_port, loop=loop, state=state.state)
        for community in state.config.snmp.trap.require_community:
            trap_receiver.add_community(community)
        trap_receiver.auto_subscribe_observers()

        try:
            loop.run_until_complete(trap_receiver.open())
        except PermissionError:
            _log.fatal(
                "Permission denied on UDP port %s. Use --trap-port to specify unprivileged port, or run as root",
                args.trap_port,
            )
            sys.exit(errno.EACCES)
    if args.user:
        switch_to_user(args.user)

    setup_initial_job_schedule(loop, args)

    server = ZinoServer(loop=loop, state=state.state, polldevs=state.polldevs, config=state.config)
    server.serve()

    # Start SNMP agent if enabled
    if state.config.snmp.agent.enabled:
        snmp_agent = ZinoSnmpAgent(
            listen_address=state.config.snmp.agent.address,
            listen_port=state.config.snmp.agent.port,
            community=state.config.snmp.agent.community,
        )
        # Schedule the SNMP agent to start as a coroutine
        loop.create_task(snmp_agent.open())

    try:
        loop.run_forever()
    except (KeyboardInterrupt, SystemExit):
        pass

    return True


def setup_initial_job_schedule(loop: AbstractEventLoop, args: argparse.Namespace) -> None:
    """Schedules all recurring and single-run jobs"""
    scheduler = get_scheduler()
    scheduler.start()

    # Register job tracker to monitor running jobs
    job_tracker = get_job_tracker()
    job_tracker.register_with_scheduler(scheduler)
    job_tracker.setup_signal_handler(loop)

    scheduler.add_job(
        func=load_and_schedule_polldevs,
        id="load_and_schedule_polldevs",
        trigger="interval",
        args=(state.config.polling.file,),
        minutes=state.config.polling.period,
        next_run_time=datetime.now(),
    )
    # Schedule state dumping as often as configured in
    # 'config.persistence.period' and reschedule whenever events are committed
    scheduler.add_job(
        func=state.state.dump_state_to_file,
        trigger="interval",
        args=(state.config.persistence.file,),
        id=STATE_DUMP_JOB_ID,
        minutes=state.config.persistence.period,
    )
    # Schedule planned maintenance
    scheduler.add_job(
        func=state.state.planned_maintenances.update_pm_states,
        id="update_pm_states",
        trigger="interval",
        args=(state.state,),
        minutes=1,
        next_run_time=datetime.now(),
    )
    # Schedule periodic flap statistics aging
    scheduler.add_job(
        func=flaps.age_flapping_states,
        id="age_flapping_states",
        args=(state.state, state.polldevs),
        trigger="interval",
        seconds=flaps.FLAP_DECREMENT_INTERVAL_SECONDS,
        next_run_time=datetime.now(),
    )
    state.state.events.add_event_observer(reschedule_dump_state_on_commit)
    state.state.planned_maintenances.add_pm_observer(reschedule_dump_state_on_pm_change)

    # Schedule removing events that have been closed for a certain time
    scheduler.add_job(
        func=state.state.events.delete_expired_events,
        id="delete_expired_events",
        trigger="interval",
        minutes=30,
    )

    scheduler.add_job(
        func=log_snmp_session_stats,
        id="log_snmp_session_stats",
        trigger="interval",
        minutes=1,
    )

    if args.stop_in:
        _log.info("Instructed to stop in %s seconds", args.stop_in)
        scheduler.add_job(
            func=loop.stop,
            trigger="date",
            run_date=datetime.now() + timedelta(seconds=args.stop_in),
            name="Stop server on timeout",
            id="stop-process",
        )


def switch_to_user(username: str):
    """Switch the process to another user (aka. drop privileges)"""

    # Get UID/GID of current user
    old_uid = os.getuid()
    old_gid = os.getgid()

    try:
        # Try to get information about the given username
        user = pwd.getpwnam(username)
    except KeyError:
        _log.error("Could not find user %s", username)
        return False

    if old_uid == user.pw_uid:
        # Already running as the given user
        _log.debug("Already running as uid/gid %d/%d.", old_uid, old_gid)
        return True

    try:
        # Set primary group
        os.setgid(user.pw_gid)

        # Set non-primary groups
        gids = [g.gr_gid for g in grp.getgrall() if username in g.gr_mem]
        if gids:
            os.setgroups(gids)

        # Set user id
        os.setuid(user.pw_uid)
    except OSError as error:
        # Failed changing uid/gid
        _log.error(
            "Failed changing uid/gid from %d/%d to %d/%d (%s)", old_uid, old_gid, user.pw_uid, user.pw_gid, error
        )
        return False

    # Switch successful
    _log.info("Dropped privileges to user %s", username)
    _log.debug("uid/gid changed from %d/%d to %d/%d.", old_uid, old_gid, user.pw_uid, user.pw_gid)
    return True


def reschedule_dump_state_on_commit(new_event: Event, old_event: Optional[Event] = None) -> None:
    """Observer that reschedules the state dumper job whenever an event is committed and there's
    more than `MINIMUM_STATE_DUMP_INTERVAL` time until the next scheduled state dump.
    """

    reschedule_dump_state(f"event {new_event.id} committed")


def reschedule_dump_state_on_pm_change() -> None:
    """Observer that reschedules the state dumper job whenever the planned maintenance
    dict is changed and there's more than `MINIMUM_STATE_DUMP_INTERVAL` time until the next scheduled
    state dump.
    """
    reschedule_dump_state("planned maintenances changed")


def reschedule_dump_state(log_msg: str) -> None:
    scheduler = get_scheduler()
    job = scheduler.get_job(job_id=STATE_DUMP_JOB_ID)
    next_run = datetime.now(tz=tzlocal.get_localzone()) + MINIMUM_STATE_DUMP_INTERVAL
    if job.next_run_time > next_run:
        _log.debug("%s, Rescheduling state dump from %s to %s", log_msg, job.next_run_time, next_run)
        job.modify(next_run_time=next_run)


def log_snmp_session_stats():
    """Logs debug information about the current number of SNMP session objects"""
    logger = logging.getLogger("zino.snmp")
    if not logger.isEnabledFor(logging.DEBUG):
        return  # Skip potentially expensive count operations if debug logging is not enabled

    backend = import_snmp_backend()
    log_low_level = "netsnmpy" in backend.__name__

    from zino.snmp import SNMP, _snmp_sessions

    device_count = len(state.state.devices)
    reusable = len(_snmp_sessions) if _snmp_sessions else 0

    msg = "(SNMP) session update: routers=%d, reusable (zino)=%d, gc reachable (high-level)=%d"
    if log_low_level:
        from netsnmpy.session import Session

        counts = _count_reachable_objects(SNMP, Session)
        counts = counts[SNMP], counts[Session]
        msg += ", gc reachable (low-level)=%d"
    else:
        counts = _count_reachable_objects(SNMP)
        counts = (counts[SNMP],)

    logger.debug(msg, device_count, reusable, *counts)


def _count_reachable_objects(*types):
    """Returns the number of objects of the given types that are currently reachable by the garbage collector.

    Designed to do a single pass over all objects in memory, for some semblance of efficiency.
    """
    counts = defaultdict(int)
    for obj in gc.get_objects():
        for t in types:
            if isinstance(obj, t):
                counts[t] += 1
    return counts


def parse_args(arguments=None):
    parser = argparse.ArgumentParser(description="Zino is not OpenView")
    parser.add_argument(
        "--polldevs",
        type=str,
        required=False,
        help="Path to the pollfile",
    )
    parser.add_argument(
        "--config-file",
        type=str,
        required=False,
        help="Path to zino configuration file",
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        default=False,
        help="Set global log level to DEBUG. Very verbose! For more fine-grained control of logging configuration, it "
        "is recommended to use the 'logging' section in the configuration file.",
    )
    parser.add_argument("--stop-in", type=int, default=None, help="Stop zino after N seconds.", metavar="N")
    parser.add_argument(
        "--trap-port",
        type=int,
        metavar="PORT",
        default=162,
        help="Which UDP port to listen for traps on.  Default value is 162.  Any value below 1024 requires root "
        "privileges.  Setting to 0 disables SNMP trap monitoring.",
    )
    parser.add_argument(
        "--user", metavar="USER", help="Switch to this user immediately after binding to privileged ports"
    )
    args = parser.parse_args(args=arguments)
    return args


if __name__ == "__main__":
    main()



from datetime import timedelta
from unittest.mock import Mock

import pytest

from zino.api.notify import Notification, Zino1NotificationProtocol
from zino.api.server import ZinoServer
from zino.state import ZinoState
from zino.statemodels import EventState, ReachabilityEvent
from zino.time import now


class TestZino1NotificationProtocol:
    def test_init_should_succeed(self):
        assert Zino1NotificationProtocol()

    def test_when_not_connected_then_peer_name_should_be_none(self):
        protocol = Zino1NotificationProtocol()
        assert protocol.peer_name is None

    def test_when_connected_then_peer_name_should_be_available(self):
        expected = "foobar"
        protocol = Zino1NotificationProtocol()
        fake_transport = Mock()
        fake_transport.get_extra_info.return_value = expected
        protocol.connection_made(fake_transport)

        assert protocol.peer_name == expected

    def test_when_connected_it_should_register_instance_in_server(self, event_loop):
        server = ZinoServer(loop=event_loop, state=ZinoState(), polldevs=dict())
        protocol = Zino1NotificationProtocol(server=server)
        fake_transport = Mock()
        protocol.connection_made(fake_transport)

        assert protocol.nonce in server.notification_channels
        assert server.notification_channels[protocol.nonce] is protocol

    def test_when_disconnected_it_should_deregister_instance_from_server(self, event_loop):
        server = ZinoServer(loop=event_loop, state=ZinoState(), polldevs=dict())
        protocol = Zino1NotificationProtocol(server=server)
        fake_transport = Mock()
        protocol.connection_made(fake_transport)
        protocol.connection_lost(exc=None)

        assert protocol.nonce not in server.notification_channels

    def test_notify_should_output_text_line(self):
        protocol = Zino1NotificationProtocol()
        fake_transport = Mock()
        protocol.connection_made(fake_transport)
        protocol.notify(Notification(42, "test", "data"))
        assert fake_transport.write.called
        response = fake_transport.write.call_args[0][0]
        assert response.startswith(b"42 test data")

    def test_tied_to_should_be_settable_and_gettable(self, event_loop):
        server = ZinoServer(loop=event_loop, state=ZinoState(), polldevs=dict())
        protocol = Zino1NotificationProtocol()

        protocol.tied_to = server
        assert protocol.tied_to == server

    def test_goodbye_should_close_transport(self, event_loop):
        server = ZinoServer(loop=event_loop, state=ZinoState(), polldevs=dict())
        protocol = Zino1NotificationProtocol(server=server)
        fake_transport = Mock()
        protocol.connection_made(fake_transport)
        protocol.goodbye()
        assert fake_transport.close.called


class TestZino1NotificationProtocolBuildNotifications:
    def test_should_make_notifications_for_regular_changed_attrs(self, fake_event):
        protocol = Zino1NotificationProtocol()
        protocol._state.events.events[fake_event.id] = fake_event
        event_copy = fake_event.model_copy(deep=True)
        event_copy.updated = now() + timedelta(seconds=5)

        notifications = list(
            protocol.build_notifications(state=protocol._state, new_event=event_copy, old_event=fake_event)
        )
        assert notifications == [Notification(event_id=42, change_type="attr", value="updated")]

    def test_should_make_notifications_for_log_changes(self, fake_event):
        protocol = Zino1NotificationProtocol()
        protocol._state.events.events[fake_event.id] = fake_event
        event_copy = fake_event.model_copy(deep=True)
        event_copy.add_log("foo")

        notifications = list(
            protocol.build_notifications(state=protocol._state, new_event=event_copy, old_event=fake_event)
        )
        assert Notification(event_id=42, change_type="log", value=1) in notifications

    def test_should_make_notifications_for_history_changes(self, fake_event):
        protocol = Zino1NotificationProtocol()
        protocol._state.events.events[fake_event.id] = fake_event
        event_copy = fake_event.model_copy(deep=True)
        event_copy.add_history("foo")

        notifications = list(
            protocol.build_notifications(state=protocol._state, new_event=event_copy, old_event=fake_event)
        )
        assert Notification(event_id=42, change_type="history", value=1) in notifications

    def test_should_make_notifications_for_state_changes(self, fake_event):
        protocol = Zino1NotificationProtocol()
        protocol._state.events.events[fake_event.id] = fake_event
        event_copy = fake_event.model_copy(deep=True)
        event_copy.state = EventState.IGNORED

        old, new = fake_event.state.value, event_copy.state.value
        expected = f"{old} {new}"

        notifications = list(
            protocol.build_notifications(state=protocol._state, new_event=event_copy, old_event=fake_event)
        )
        assert Notification(event_id=42, change_type="state", value=expected) in notifications

    def test_when_event_is_scavenged_it_should_only_produce_a_single_final_notification(self, fake_event):
        protocol = Zino1NotificationProtocol()
        fake_event.state = EventState.CLOSED
        notifications = list(protocol.build_notifications(state=protocol._state, new_event=fake_event, old_event=None))
        assert Notification(event_id=fake_event.id, change_type="scavenged", value=None) in notifications
        assert len(notifications) == 1


class TestZino1NotificationProtocolBuildAndSendNotifications:
    def test_should_send_notifications_only_to_tied_channels(self, event_loop, fake_event, changed_fake_event):
        server = ZinoServer(loop=event_loop, state=ZinoState(), polldevs=dict())
        channel1 = Mock()
        channel2 = Mock()
        mock_api = Mock()

        channel1.tied_to = mock_api
        channel2.tied_to = None

        server.notification_channels["a"] = channel1
        server.notification_channels["b"] = channel2

        Zino1NotificationProtocol.build_and_send_notifications(
            server, new_event=changed_fake_event, old_event=fake_event
        )
        assert channel1.notify.called
        assert not channel2.notify.called


@pytest.fixture
def fake_event() -> ReachabilityEvent:
    return ReachabilityEvent(id=42, router="example-gw.example.org", state=EventState.OPEN)


@pytest.fixture
def changed_fake_event(fake_event) -> ReachabilityEvent:
    copy = fake_event.model_copy(deep=True)
    copy.add_history("this fake event has been changed")
    return copy



from unittest.mock import patch

import pytest

from zino.config.models import Configuration, PollDevice
from zino.state import ZinoState
from zino.tasks.reachabletask import ReachableTask


class TestTask:
    async def test_get_sysuptime_returns_uptime(self, snmpsim, snmp_test_port):
        device = PollDevice(
            name="buick.lab.example.org",
            address="127.0.0.1",
            community="public",
            port=snmp_test_port,
        )
        state = ZinoState()
        task = ReachableTask(device, state, Configuration())
        uptime = await task._get_uptime()
        assert uptime

    async def test_get_sysuptime_raises_timeout_error(self):
        device = PollDevice(name="nonexist", address="127.0.0.1", community="invalid", port=666)
        state = ZinoState()
        task = ReachableTask(device, state, Configuration())
        with patch("zino.snmp.SNMP.get") as get_mock:
            get_mock.side_effect = TimeoutError
            with pytest.raises(TimeoutError):
                await task._get_uptime()
